import json
from random import randint
import re
import copy


import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
import requests


from solarforecastarbiter.io import api, utils
from solarforecastarbiter import datamodel


def test_request_cli_access_token_mocked(requests_mock):
    requests_mock.register_uri(
        'POST', 'https://solarforecastarbiter.auth0.com/oauth/token',
        content=b'{"access_token": "token"}')
    assert api.request_cli_access_token('test', 'pass') == 'token'


def test_request_cli_access_token_real():
    try:
        requests.get('https://solarforecastarbiter.auth0.com')
    except Exception:  # pragma: no cover
        return pytest.skip('Cannot connect to Auth0')
    else:
        assert api.request_cli_access_token('testing@solarforecastarbiter.org',
                                            'Thepassword123!') is not None


def test_request_cli_access_token_mocked_kwargs(requests_mock):
    m = requests_mock.register_uri(
        'POST', 'https://solarforecastarbiter.auth0.com/oauth/token',
        content=b'{"access_token": "token"}')
    token = api.request_cli_access_token(
        'test', 'pass', cert="some_certificate.crt")
    assert m.request_history[0].cert == "some_certificate.crt"
    assert token == 'token'


def test_apisession_init(requests_mock):
    session = api.APISession('TOKEN')
    requests_mock.register_uri('GET', session.base_url)
    res = session.get('')
    assert res.request.headers['Authorization'] == 'Bearer TOKEN'


def test_apisession_init_hidden(requests_mock):
    ht = utils.HiddenToken('TOKEN')
    session = api.APISession(ht)
    requests_mock.register_uri('GET', session.base_url)
    res = session.get('')
    assert res.request.headers['Authorization'] == 'Bearer TOKEN'


@pytest.mark.parametrize('method,endpoint,expected', [
    ('GET', '/', 'https://api.solarforecastarbiter.org/'),
    ('GET', '', 'https://api.solarforecastarbiter.org/'),
    ('GET', '/sites/', 'https://api.solarforecastarbiter.org/sites/'),
    ('GET', 'sites/', 'https://api.solarforecastarbiter.org/sites/'),
    ('POST', '/observations/obsid/values',
     'https://api.solarforecastarbiter.org/observations/obsid/values'),
])
def test_apisession_request(endpoint, method, expected, requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri(method, matcher)
    res = session.request(method, endpoint)
    assert res.request.url == expected


@pytest.fixture()
def mock_get_site(requests_mock, site_text, many_sites_text):
    def get_site_from_text(request, context):
        site_id = request.url.split('/')[-1]
        if site_id == '':
            return many_sites_text
        else:
            sites = json.loads(many_sites_text)
            for site in sites:
                if site['site_id'] == site_id:
                    return json.dumps(site).encode('utf-8')

    matcher = re.compile(f'https://api.solarforecastarbiter.org/sites/.*')
    requests_mock.register_uri('GET', matcher, content=get_site_from_text)


@pytest.fixture()
def mock_get_observation(requests_mock, many_observations_text,
                         mock_get_site):
    def get_observation_from_text(request, context):
        obs_id = request.url.split('/')[-2]
        if obs_id == '':  # pragma: no cover
            return many_observations_text
        else:
            obs = json.loads(many_observations_text)
            for ob in obs:
                if ob['observation_id'] == obs_id:
                    return json.dumps(ob).encode('utf-8')

    matcher = re.compile(
        'https://api.solarforecastarbiter.org/observations/.*/metadata')
    requests_mock.register_uri(
        'GET', matcher, content=get_observation_from_text)


@pytest.fixture()
def mock_get_agg(requests_mock, aggregate_text, mock_get_observation):
    matcher = re.compile(f'https://api.solarforecastarbiter.org/aggregates/.*')
    requests_mock.register_uri('GET', matcher, content=aggregate_text)


@pytest.fixture()
def mock_list_sites(mocker, many_sites):
    mocker.patch('solarforecastarbiter.io.api.APISession.list_sites',
                 return_value=many_sites)


def test_apisession_get_site(mock_get_site, get_site):
    session = api.APISession('')
    site = session.get_site('123e4567-e89b-12d3-a456-426655440002')
    assert site == get_site('123e4567-e89b-12d3-a456-426655440002')


def test_apisession_get_site_dne(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri('GET', matcher, status_code=404)
    with pytest.raises(requests.exceptions.HTTPError):
        session.get_site('123e4567-e89b-12d3-a456-426655440002')


def test_apisession_list_sites(requests_mock, many_sites_text, many_sites):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri('GET', matcher, content=many_sites_text)
    site_list = session.list_sites()
    assert site_list == many_sites


def test_apisession_list_sites_in_zone(requests_mock, many_sites_text,
                                       many_sites):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri('GET', matcher, content=json.dumps(
        json.loads(many_sites_text)[1:]).encode())
    site_list = session.list_sites_in_zone('Reference Region 5')
    assert site_list == many_sites[1:]


def test_apisession_list_sites_empty(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri('GET', matcher, content=b"[]")
    site_list = session.list_sites()
    assert site_list == []


def test_apisession_list_sites_in_zone_empty(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri('GET', matcher, content=b"[]")
    site_list = session.list_sites_in_zone('bad zone')
    assert site_list == []


def test_apisession_search_climatezones(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/.*')
    requests_mock.register_uri(
        'GET', matcher,
        content=b'[{"name": "Reference Region 3", "created_at": 0}]')
    zone_list = session.search_climatezones(32.1, -110.8)
    assert zone_list == ["Reference Region 3"]


def test_apisession_create_site(requests_mock, single_site, site_text):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/sites/.*')
    requests_mock.register_uri('POST', matcher,
                               text=single_site.site_id)
    requests_mock.register_uri('GET', matcher, content=site_text)
    site_dict = single_site.to_dict()
    del site_dict['site_id']
    del site_dict['provider']
    del site_dict['extra_parameters']
    ss = type(single_site).from_dict(site_dict)
    new_site = session.create_site(ss)
    assert new_site == single_site


def test_apisession_get_observation(requests_mock, single_observation,
                                    single_observation_text, mock_get_site):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('GET', matcher, content=single_observation_text)
    obs = session.get_observation('')
    assert obs == single_observation


def test_apisession_list_observations(requests_mock, many_observations,
                                      many_observations_text, mock_list_sites):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('GET', matcher, content=many_observations_text)
    obs_list = session.list_observations()
    assert obs_list == many_observations


def test_apisession_list_observation_single(requests_mock, single_observation,
                                            single_observation_text,
                                            mock_list_sites):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('GET', matcher, content=single_observation_text)
    obs_list = session.list_observations()
    assert obs_list == [single_observation]


def test_apisession_list_observations_empty(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('GET', matcher, content=b"[]")
    obs_list = session.list_observations()
    assert obs_list == []


def test_apisession_create_observation(requests_mock, single_observation,
                                       single_observation_text, mock_get_site):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('POST', matcher,
                               text=single_observation.observation_id)
    requests_mock.register_uri('GET', matcher, content=single_observation_text)
    observation_dict = single_observation.to_dict()
    del observation_dict['observation_id']
    del observation_dict['extra_parameters']
    ss = type(single_observation).from_dict(observation_dict)
    new_observation = session.create_observation(ss)
    assert new_observation == single_observation


def test_apisession_get_forecast(requests_mock, single_forecast,
                                 single_forecast_text, mock_get_site):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/.*')
    requests_mock.register_uri('GET', matcher, content=single_forecast_text)
    fx = session.get_forecast('')
    assert fx == single_forecast


def test_apisession_list_forecasts(requests_mock, many_forecasts,
                                   many_forecasts_text, mock_get_site,
                                   mock_get_agg):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/.*')
    requests_mock.register_uri('GET', matcher, content=many_forecasts_text)
    fx_list = session.list_forecasts()
    assert fx_list == many_forecasts


def test_apisession_list_forecast_single(requests_mock, single_forecast,
                                         single_forecast_text, mock_get_site,
                                         mock_get_agg):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/.*')
    requests_mock.register_uri('GET', matcher, content=single_forecast_text)
    fx_list = session.list_forecasts()
    assert fx_list == [single_forecast]


def test_apisession_create_forecast(requests_mock, single_forecast,
                                    single_forecast_text, mock_get_site,
                                    mock_get_agg):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/single/.*')
    requests_mock.register_uri('POST', matcher,
                               text=single_forecast.forecast_id)
    requests_mock.register_uri('GET', matcher, content=single_forecast_text)
    forecast_dict = single_forecast.to_dict()
    del forecast_dict['forecast_id']
    del forecast_dict['extra_parameters']
    ss = type(single_forecast).from_dict(forecast_dict)
    new_forecast = session.create_forecast(ss)
    assert new_forecast == single_forecast


def test_apisession_create_forecast_agg(
        requests_mock, aggregate, aggregateforecast,
        aggregate_forecast_text, mock_get_agg):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/single/.*')
    requests_mock.register_uri('POST', matcher,
                               text=aggregateforecast.forecast_id)
    requests_mock.register_uri('GET', matcher, content=aggregate_forecast_text)
    forecast_dict = aggregateforecast.to_dict()
    del forecast_dict['forecast_id']
    del forecast_dict['extra_parameters']
    ss = type(aggregateforecast).from_dict(forecast_dict)
    new_forecast = session.create_forecast(ss)
    assert new_forecast == aggregateforecast


def test_apisession_list_forecasts_empty(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/.*')
    requests_mock.register_uri('GET', matcher, content=b'[]')
    fx_list = session.list_forecasts()
    assert fx_list == []


def test_apisession_get_prob_forecast(requests_mock, prob_forecasts,
                                      prob_forecast_text, mock_get_site,
                                      prob_forecast_constant_value_text):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    matcher = re.compile(session.base_url + r'/forecasts/cdf/[\w-]*$')
    requests_mock.register_uri('GET', matcher, content=prob_forecast_text)
    fx = session.get_probabilistic_forecast('')
    assert fx == prob_forecasts


def test_apisession_list_prob_forecasts(requests_mock, many_prob_forecasts,
                                        many_prob_forecasts_text,
                                        mock_list_sites, mock_get_site,
                                        prob_forecast_constant_value_text,
                                        mock_get_agg):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    matcher = re.compile(session.base_url + r'/forecasts/cdf/$')
    requests_mock.register_uri(
        'GET', matcher, content=many_prob_forecasts_text)
    fx_list = session.list_probabilistic_forecasts()
    assert fx_list == many_prob_forecasts


def test_apisession_list_prob_forecast_single(
        requests_mock, prob_forecasts, prob_forecast_text, mock_list_sites,
        mock_get_site, prob_forecast_constant_value_text, mock_get_agg):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    matcher = re.compile(session.base_url + r'/forecasts/cdf/$')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_text)
    fx_list = session.list_probabilistic_forecasts()
    assert fx_list == [prob_forecasts]


def test_apisession_list_prob_forecasts_empty(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/$')
    requests_mock.register_uri('GET', matcher, content=b'[]')
    fx_list = session.list_probabilistic_forecasts()
    assert fx_list == []


def test_apisession_get_prob_forecast_constant_value(
        requests_mock, prob_forecast_constant_value,
        prob_forecast_constant_value_text, mock_get_site):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    fx = session.get_probabilistic_forecast_constant_value('')
    assert fx == prob_forecast_constant_value


def test_apisession_get_prob_forecast_constant_value_site(
        requests_mock, prob_forecast_constant_value,
        prob_forecast_constant_value_text, single_site):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    fx = session.get_probabilistic_forecast_constant_value(
        '', site=single_site)
    assert fx == prob_forecast_constant_value


def test_apisession_get_agg_prob_forecast_constant_value(
        requests_mock, agg_prob_forecast_constant_value,
        agg_prob_forecast_constant_value_text, mock_get_agg,
        aggregate):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=agg_prob_forecast_constant_value_text)
    fx = session.get_probabilistic_forecast_constant_value('')
    assert fx == agg_prob_forecast_constant_value


def test_apisession_get_agg_prob_forecast_constant_value_agg(
        requests_mock, agg_prob_forecast_constant_value,
        agg_prob_forecast_constant_value_text, mock_get_agg,
        aggregate):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=agg_prob_forecast_constant_value_text)
    fx = session.get_probabilistic_forecast_constant_value(
        '', aggregate=aggregate)
    assert fx == agg_prob_forecast_constant_value


def test_apisession_get_prob_forecast_constant_value_site_error(
        requests_mock, prob_forecast_constant_value_text, single_site):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    site_dict = single_site.to_dict()
    site_dict['site_id'] = 'nope'
    site = datamodel.Site.from_dict(site_dict)
    with pytest.raises(ValueError):
        session.get_probabilistic_forecast_constant_value('', site=site)


def test_apisession_get_prob_forecast_constant_value_agg_error(
        requests_mock, agg_prob_forecast_constant_value_text, aggregate):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=agg_prob_forecast_constant_value_text)
    agg_dict = aggregate.to_dict()
    agg_dict['aggregate_id'] = 'nope'
    agg = datamodel.Aggregate.from_dict(agg_dict)
    with pytest.raises(ValueError):
        session.get_probabilistic_forecast_constant_value(
            '', aggregate=agg)


def test_apisession_create_prob_forecast(requests_mock, prob_forecasts,
                                         prob_forecast_text, mock_get_site,
                                         prob_forecast_constant_value_text):
    session = api.APISession('')
    matcher = re.compile(session.base_url + r'/forecasts/cdf/$')
    requests_mock.register_uri('POST', matcher,
                               text=prob_forecasts.forecast_id)
    matcher = re.compile(
        f'{session.base_url}/forecasts/cdf/{prob_forecasts.forecast_id}$')
    requests_mock.register_uri('GET', matcher, content=prob_forecast_text)
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=prob_forecast_constant_value_text)
    forecast_dict = prob_forecasts.to_dict()
    del forecast_dict['forecast_id']
    del forecast_dict['extra_parameters']
    ss = type(prob_forecasts).from_dict(forecast_dict)
    new_forecast = session.create_probabilistic_forecast(ss)
    assert new_forecast == prob_forecasts


def test_apisession_create_prob_forecast_agg(
        requests_mock, aggregate_prob_forecast,
        aggregate_prob_forecast_text, mock_get_agg,
        agg_prob_forecast_constant_value_text):
    session = api.APISession('')
    matcher = re.compile(session.base_url + r'/forecasts/cdf/$')
    requests_mock.register_uri('POST', matcher,
                               text=aggregate_prob_forecast.forecast_id)
    matcher = re.compile(
        f'{session.base_url}/forecasts/cdf/{aggregate_prob_forecast.forecast_id}$')  # NOQA
    requests_mock.register_uri(
        'GET', matcher, content=aggregate_prob_forecast_text)
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*')
    requests_mock.register_uri(
        'GET', matcher, content=agg_prob_forecast_constant_value_text)
    forecast_dict = aggregate_prob_forecast.to_dict()
    del forecast_dict['forecast_id']
    del forecast_dict['extra_parameters']
    ss = type(aggregate_prob_forecast).from_dict(forecast_dict)
    new_forecast = session.create_probabilistic_forecast(ss)
    assert new_forecast == aggregate_prob_forecast


@pytest.fixture(params=[0, 1])
def obs_start_end(request):
    if request.param == 0:
        return (pd.Timestamp('2019-01-01T12:00:00-0700'),
                pd.Timestamp('2019-01-01T12:25:00-0700'))
    else:
        return ('2019-01-01T12:00:00-0700',
                '2019-01-01T12:25:00-0700')


@pytest.mark.parametrize('func', ['get_observation_values',
                                  'get_aggregate_values'])
def test_apisession_get_obs_agg_values(
        requests_mock, observation_values, observation_values_text,
        obs_start_end, func):
    session = api.APISession('')
    matcher = re.compile(
        f'{session.base_url}/(observations|aggregates)/.*/values')
    requests_mock.register_uri('GET', matcher, content=observation_values_text)
    out = getattr(session, func)('obsid', *obs_start_end)
    pdt.assert_frame_equal(out, observation_values)


@pytest.mark.parametrize('label,theslice', [
    (None, slice(0, 10)),
    ('beginning', slice(0, -1)),
    ('ending', slice(1, 10))
])
@pytest.mark.parametrize('func', ['get_observation_values',
                                  'get_aggregate_values'])
def test_apisession_get_obs_agg_values_interval_label(
        requests_mock, observation_values, observation_values_text,
        label, theslice, obs_start_end, func):
    session = api.APISession('')
    matcher = re.compile(
        f'{session.base_url}/(observations|aggregates)/.*/values')
    requests_mock.register_uri('GET', matcher, content=observation_values_text)
    out = getattr(session, func)(
        'obsid', obs_start_end[0], obs_start_end[1], label)
    pdt.assert_frame_equal(out, observation_values.iloc[theslice])


@pytest.fixture()
def empty_df():
    return pd.DataFrame(
        [], columns=['value', 'quality_flag'],
        index=pd.DatetimeIndex([], name='timestamp', tz='UTC')
    ).astype({'value': float, 'quality_flag': int})


@pytest.mark.parametrize('func', ['get_observation_values',
                                  'get_aggregate_values'])
def test_apisession_get_obs_agg_values_empty(requests_mock, empty_df,
                                             func):
    session = api.APISession('')
    matcher = re.compile(
        f'{session.base_url}/(observations|aggregates)/.*/values')
    requests_mock.register_uri('GET', matcher, content=b'{"values":[]}')
    out = getattr(session, func)(
        'obsid', pd.Timestamp('2019-01-01T12:00:00-0700'),
        pd.Timestamp('2019-01-01T12:25:00-0700'))
    pdt.assert_frame_equal(out, empty_df)


@pytest.fixture(params=[0, 1])
def fx_start_end(request):
    if request.param == 0:
        return (pd.Timestamp('2019-01-01T06:00:00-0700'),
                pd.Timestamp('2019-01-01T11:00:00-0700'))
    else:
        return ('2019-01-01T06:00:00-0700',
                '2019-01-01T11:00:00-0700')


def test_apisession_get_forecast_values(requests_mock, forecast_values,
                                        forecast_values_text, fx_start_end):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/single/.*/values')
    requests_mock.register_uri('GET', matcher, content=forecast_values_text)
    out = session.get_forecast_values(
        'fxid', *fx_start_end)
    pdt.assert_series_equal(out, forecast_values)


def test_apisession_get_prob_forecast_constant_value_values(
        requests_mock, forecast_values, forecast_values_text, fx_start_end):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*/values')
    requests_mock.register_uri('GET', matcher, content=forecast_values_text)
    out = session.get_probabilistic_forecast_constant_value_values(
        'fxid', *fx_start_end)
    pdt.assert_series_equal(out, forecast_values)


@pytest.mark.parametrize('col_values', [
    ['25.0', '50.0', '75.0'], [25., 50., 75.]])
def test_apisession_get_prob_forecast_values(
        requests_mock, mock_get_site, col_values,
        prob_forecast_values, prob_forecast_values_text_list,
        prob_forecast_text, prob_forecast_constant_value_text, fx_start_end):
    session = api.APISession('')

    # modify ProbabilisticForecast json text
    probfx_json = json.loads(prob_forecast_text)
    probfx_json['forecast_id'] = 'FXID'
    probfx_json['axis'] = 'y'
    cvs_json = json.loads(prob_forecast_constant_value_text)
    cvs = []
    for col in col_values:
        new_cv = copy.deepcopy(cvs_json)
        new_cv['constant_value'] = col
        # fixture has keys like CV25
        # we need to convert either float 25. or str '25.0' to int 25
        # str(25.) == '25.0', int('25.0') raises ValueError, so we have
        # to chain it all together
        new_cv['forecast_id'] = 'CV' + str(int(float(col)))
        new_cv['axis'] = 'y'
        cvs.append(new_cv)
    probfx_json['constant_values'] = cvs

    matcher = re.compile(session.base_url + r'/forecasts/cdf/[\w-]*$')
    requests_mock.register_uri(
        'GET', matcher, content=json.dumps(probfx_json).encode('utf-8'))
    for cv in cvs:
        id = cv['forecast_id']
        cv = json.dumps(copy.deepcopy(cv))
        requests_mock.register_uri(
            'GET', f'/forecasts/cdf/single/{id}', content=cv.encode('utf-8'))
    for cvv in prob_forecast_values_text_list:
        id = json.loads(cvv)['forecast_id']
        requests_mock.register_uri(
            'GET', f'/forecasts/cdf/single/{id}/values', content=cvv)
    out = session.get_probabilistic_forecast_values('', *fx_start_end)
    pdt.assert_frame_equal(out, prob_forecast_values)


@pytest.mark.parametrize('label,theslice', [
    (None, slice(0, 10)),
    ('beginning', slice(0, -1)),
    ('ending', slice(1, 10))
])
def test_apisession_get_forecast_values_interval_label(
        requests_mock, forecast_values, forecast_values_text, label, theslice,
        fx_start_end):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/single/.*/values')
    requests_mock.register_uri('GET', matcher, content=forecast_values_text)
    out = session.get_forecast_values(
        'fxid', fx_start_end[0], fx_start_end[1], label)
    pdt.assert_series_equal(out, forecast_values.iloc[theslice])


def test_apisession_get_forecast_values_empty(requests_mock, empty_df):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/single/.*/values')
    requests_mock.register_uri('GET', matcher, content=b'{"values":[]}')
    out = session.get_forecast_values(
        'fxid', pd.Timestamp('2019-01-01T06:00:00-0700'),
        pd.Timestamp('2019-01-01T11:00:00-0700'))
    pdt.assert_series_equal(out, empty_df['value'])


def test_apisession_post_observation_values(requests_mock, observation_values):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/observations/.*/values')
    mocked = requests_mock.register_uri('POST', matcher)
    session.post_observation_values('obsid', observation_values)
    # observation_values_text has a different timestamp format
    assert mocked.request_history[0].text == '{"values":[{"timestamp":"2019-01-01T19:00:00Z","value":0.0,"quality_flag":0},{"timestamp":"2019-01-01T19:05:00Z","value":1.0,"quality_flag":0},{"timestamp":"2019-01-01T19:10:00Z","value":1.5,"quality_flag":0},{"timestamp":"2019-01-01T19:15:00Z","value":9.9,"quality_flag":1},{"timestamp":"2019-01-01T19:20:00Z","value":2.0,"quality_flag":0},{"timestamp":"2019-01-01T19:25:00Z","value":-999.0,"quality_flag":3}]}'  # NOQA


def test_apisession_post_forecast_values(requests_mock, forecast_values):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/single/.*/values')
    mocked = requests_mock.register_uri('POST', matcher)
    session.post_forecast_values('fxid', forecast_values)
    assert mocked.request_history[0].text == '{"values":[{"timestamp":"2019-01-01T13:00:00Z","value":0.0},{"timestamp":"2019-01-01T14:00:00Z","value":1.0},{"timestamp":"2019-01-01T15:00:00Z","value":2.0},{"timestamp":"2019-01-01T16:00:00Z","value":3.0},{"timestamp":"2019-01-01T17:00:00Z","value":4.0},{"timestamp":"2019-01-01T18:00:00Z","value":5.0}]}'  # NOQA


def test_apisession_post_prob_forecast_constant_value_values(
        requests_mock, forecast_values):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/forecasts/cdf/single/.*/values')
    mocked = requests_mock.register_uri('POST', matcher)
    session.post_probabilistic_forecast_constant_value_values(
        'fxid', forecast_values)
    assert mocked.request_history[0].text == '{"values":[{"timestamp":"2019-01-01T13:00:00Z","value":0.0},{"timestamp":"2019-01-01T14:00:00Z","value":1.0},{"timestamp":"2019-01-01T15:00:00Z","value":2.0},{"timestamp":"2019-01-01T16:00:00Z","value":3.0},{"timestamp":"2019-01-01T17:00:00Z","value":4.0},{"timestamp":"2019-01-01T18:00:00Z","value":5.0}]}'  # NOQA


@pytest.mark.parametrize('match, meth', [
    ('observations', 'get_observation_values'),
    ('forecasts/single', 'get_forecast_values'),
    ('forecasts/cdf', 'get_probabilistic_forecast_values'),
    ('forecasts/cdf/single',
     'get_probabilistic_forecast_constant_value_values'),
    ('aggregates', 'get_aggregate_values'),
])
def test_apisession_get_values(
        requests_mock, mocker, single_observation, single_forecast,
        prob_forecasts, prob_forecast_constant_value, aggregate, match, meth,
        fx_start_end):
    objs = {
        'observations': single_observation,
        'forecasts/single': single_forecast,
        'forecasts/cdf': prob_forecasts,
        'forecasts/cdf/single': prob_forecast_constant_value,
        'aggregates': aggregate
    }
    obj = objs[match]
    session = api.APISession('')
    matcher = re.compile(
        f'{session.base_url}/{match}/.*/values')
    requests_mock.register_uri('GET', matcher, content=b'{"values":[]}')
    status = mocker.patch(
        f'solarforecastarbiter.io.api.APISession.{meth}')
    session.get_values(obj, *fx_start_end)
    assert status.called


@pytest.fixture()
def mock_request_fxobs(report_objects, mocker):
    report, obs, fx0, fx1, agg, fxagg = report_objects
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)
    mocker.patch('solarforecastarbiter.io.api.APISession.get_aggregate',
                 return_value=agg)

    def returnone(fxid):
        if fxid == "da2bc386-8712-11e9-a1c7-0a580a8200ae":
            return fx0
        elif fxid == "49220780-76ae-4b11-bef1-7a75bdc784e3":
            return fxagg
        elif fxid == "refbc386-8712-11e9-a1c7-0a580a8200ae":
            return report.report_parameters.object_pairs[1].reference_forecast
        else:
            return fx1
    mocker.patch('solarforecastarbiter.io.api.APISession.get_forecast',
                 side_effect=returnone)


def test_apisession_get_report(requests_mock, report_text, report_objects,
                               mock_request_fxobs):
    session = api.APISession('')
    requests_mock.register_uri('GET', f'{session.base_url}/reports/',
                               content=report_text)
    out = session.get_report('')
    expected = report_objects[0]
    assert out == expected


@pytest.fixture()
def mock_request_event_fxobs(event_report_objects, mocker):
    _, obs, fx0, fx1 = event_report_objects
    mocker.patch('solarforecastarbiter.io.api.APISession.get_observation',
                 return_value=obs)

    def returnone(fxid):
        if fxid == "da2bc386-8712-11e9-a1c7-0a580a8200ae":
            return fx0
        else:
            return fx1
    mocker.patch('solarforecastarbiter.io.api.APISession.get_forecast',
                 side_effect=returnone)


def test_apisession_get_report_event(requests_mock, event_report_text,
                                     event_report_objects,
                                     mock_request_event_fxobs):
    session = api.APISession('')
    requests_mock.register_uri('GET', f'{session.base_url}/reports/',
                               content=event_report_text)
    out = session.get_report('')
    expected = event_report_objects[0]
    assert out == expected


def test_apisession_get_report_with_raw(
        requests_mock, report_text, report_objects, mock_request_fxobs,
        raw_report, mocker):
    raw = raw_report(False)
    raw_dict = raw.to_dict()
    report = json.loads(report_text)
    report['raw_report'] = raw_dict
    report_text = json.dumps(report).encode()
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_raw_report_processed_data',
        return_value=())
    session = api.APISession('')
    requests_mock.register_uri('GET', f'{session.base_url}/reports/',
                               content=report_text)
    out = session.get_report('')
    expected = report_objects[0].replace(
        raw_report=raw.replace(processed_forecasts_observations=()))
    assert out == expected


def test_apisession_list_reports(requests_mock, report_text, report_objects,
                                 mock_request_fxobs):
    session = api.APISession('')
    requests_mock.register_uri('GET', f'{session.base_url}/reports',
                               content=b'['+report_text+b']')
    out = session.list_reports()
    expected = [report_objects[0]]
    assert out == expected


def test_apisession_list_reports_single(requests_mock, report_text,
                                        report_objects, mock_request_fxobs):
    session = api.APISession('')
    requests_mock.register_uri('GET', f'{session.base_url}/reports',
                               content=report_text)
    out = session.list_reports()
    expected = [report_objects[0]]
    assert out == expected


def test_apisession_list_reports_empty(requests_mock):
    session = api.APISession('')
    requests_mock.register_uri('GET', f'{session.base_url}/reports',
                               content=b'[]')
    out = session.list_reports()
    assert out == []


def test_apisession_create_report(requests_mock, report_objects, mocker):
    session = api.APISession('')
    report = report_objects[0]
    mocked = requests_mock.register_uri('POST', f'{session.base_url}/reports/')
    mocker.patch('solarforecastarbiter.io.api.APISession.get_report',
                 return_value=report)
    expected = {
        "report_parameters": {
            "name": "NREL MIDC OASIS GHI Forecast Analysis",
            "start": "2019-04-01T00:00:00-07:00",
            "end": "2019-04-04T23:59:00-07:00",
            "forecast_fill_method": "forward",
            "filters": [
                {'quality_flags': [
                    "USER FLAGGED",
                    "NIGHTTIME",
                    "LIMITS EXCEEDED",
                    "STALE VALUES",
                    "INTERPOLATED VALUES",
                    "INCONSISTENT IRRADIANCE COMPONENTS",
                ]},
                {'time_of_day_range': ['12:00', '14:00']}],
            "metrics": ["mae", "rmse", "mbe", "s", "cost"],
            "categories": ["total", "date", "hour"],
            "object_pairs": [
                {"forecast": "da2bc386-8712-11e9-a1c7-0a580a8200ae",
                 "observation": "9f657636-7e49-11e9-b77f-0a580a8003e9",
                 "cost": "example cost",
                 "uncertainty": "1.0"},
                {"forecast": "68a1c22c-87b5-11e9-bf88-0a580a8200ae",
                 "observation": "9f657636-7e49-11e9-b77f-0a580a8003e9",
                 "normalization": "1000.0",
                 "uncertainty": "15.0",
                 "cost": "example cost",
                 "reference_forecast": "refbc386-8712-11e9-a1c7-0a580a8200ae"},
                {"forecast": "49220780-76ae-4b11-bef1-7a75bdc784e3",
                 "aggregate": "458ffc27-df0b-11e9-b622-62adb5fd6af0",
                 "cost": "example cost",
                 "uncertainty": "5.0"}
            ],
            "costs": [
                {
                    "name": "example cost",
                    "type": "constant",
                    "parameters": {
                        "cost": 1.0,
                        "aggregation": "sum",
                        "net": True
                    }
                }
            ]
        }}
    session.create_report(report)
    posted = mocked.last_request.json()
    assert posted == expected


def test_apisession_create_report_mult_costs(requests_mock, report_objects,
                                             mocker):
    session = api.APISession('')
    report = report_objects[0]
    newobj = list(report.report_parameters.object_pairs)
    ncost = datamodel.Cost(
        name='other cost',
        type='constant',
        parameters=datamodel.ConstantCost(
            cost=2.0,
            aggregation='sum',
            net=True
        )
    )
    newobj[-1] = newobj[-1].replace(cost='other cost')
    report = report.replace(
        report_parameters=report.report_parameters.replace(
            costs=(report.report_parameters.costs[0], ncost),
            object_pairs=tuple(newobj)))
    mocked = requests_mock.register_uri('POST', f'{session.base_url}/reports/')
    mocker.patch('solarforecastarbiter.io.api.APISession.get_report',
                 return_value=report)
    expected = {
        "report_parameters": {
            "name": "NREL MIDC OASIS GHI Forecast Analysis",
            "start": "2019-04-01T00:00:00-07:00",
            "end": "2019-04-04T23:59:00-07:00",
            "forecast_fill_method": "forward",
            "filters": [
                {'quality_flags': [
                    "USER FLAGGED",
                    "NIGHTTIME",
                    "LIMITS EXCEEDED",
                    "STALE VALUES",
                    "INTERPOLATED VALUES",
                    "INCONSISTENT IRRADIANCE COMPONENTS",
                ]},
                {'time_of_day_range': ['12:00', '14:00']}],
            "metrics": ["mae", "rmse", "mbe", "s", "cost"],
            "categories": ["total", "date", "hour"],
            "object_pairs": [
                {"forecast": "da2bc386-8712-11e9-a1c7-0a580a8200ae",
                 "observation": "9f657636-7e49-11e9-b77f-0a580a8003e9",
                 "cost": "example cost",
                 "uncertainty": "1.0"},
                {"forecast": "68a1c22c-87b5-11e9-bf88-0a580a8200ae",
                 "observation": "9f657636-7e49-11e9-b77f-0a580a8003e9",
                 "normalization": "1000.0",
                 "uncertainty": "15.0",
                 "cost": "example cost",
                 "reference_forecast": "refbc386-8712-11e9-a1c7-0a580a8200ae"},
                {"forecast": "49220780-76ae-4b11-bef1-7a75bdc784e3",
                 "aggregate": "458ffc27-df0b-11e9-b622-62adb5fd6af0",
                 "cost": "other cost",
                 "uncertainty": "5.0"}
            ],
            "costs": [
                {
                    "name": "example cost",
                    "type": "constant",
                    "parameters": {
                        "cost": 1.0,
                        "aggregation": "sum",
                        "net": True
                    }
                },
                {
                    "name": "other cost",
                    "type": "constant",
                    "parameters": {
                        "cost": 2.0,
                        "aggregation": "sum",
                        "net": True
                    }
                }
            ]
        }}
    session.create_report(report)
    posted = mocked.last_request.json()
    assert posted == expected


def test_apisession_create_report_no_costs(
        requests_mock, report_objects, mocker):
    session = api.APISession('')
    report = report_objects[0]
    report = report.replace(
        report_parameters=report.report_parameters.replace(
            costs=tuple(),
            object_pairs=tuple(
                op.replace(cost=None) for op in
                report.report_parameters.object_pairs)))
    mocked = requests_mock.register_uri('POST', f'{session.base_url}/reports/')
    mocker.patch('solarforecastarbiter.io.api.APISession.get_report',
                 return_value=report)
    expected = {
        "report_parameters": {
            "name": "NREL MIDC OASIS GHI Forecast Analysis",
            "start": "2019-04-01T00:00:00-07:00",
            "end": "2019-04-04T23:59:00-07:00",
            "forecast_fill_method": "forward",
            "filters": [
                {'quality_flags': [
                    "USER FLAGGED",
                    "NIGHTTIME",
                    "LIMITS EXCEEDED",
                    "STALE VALUES",
                    "INTERPOLATED VALUES",
                    "INCONSISTENT IRRADIANCE COMPONENTS",
                ]},
                {'time_of_day_range': ['12:00', '14:00']}],
            "metrics": ["mae", "rmse", "mbe", "s", "cost"],
            "categories": ["total", "date", "hour"],
            "object_pairs": [
                {"forecast": "da2bc386-8712-11e9-a1c7-0a580a8200ae",
                 "observation": "9f657636-7e49-11e9-b77f-0a580a8003e9",
                 "uncertainty": "1.0"},
                {"forecast": "68a1c22c-87b5-11e9-bf88-0a580a8200ae",
                 "observation": "9f657636-7e49-11e9-b77f-0a580a8003e9",
                 "normalization": "1000.0",
                 "uncertainty": "15.0",
                 "reference_forecast": "refbc386-8712-11e9-a1c7-0a580a8200ae"},
                {"forecast": "49220780-76ae-4b11-bef1-7a75bdc784e3",
                 "aggregate": "458ffc27-df0b-11e9-b622-62adb5fd6af0",
                 "uncertainty": "5.0"}
            ],
            "costs": []
        }}
    session.create_report(report)
    posted = mocked.last_request.json()
    assert posted == expected


def test_apisession_post_raw_report_processed_data(
        requests_mock, raw_report, report_objects, ref_forecast_id):
    _, obs, fx0, fx1, agg, fxagg = report_objects
    session = api.APISession('')
    ids = [
        fx0.forecast_id, obs.observation_id, fx1.forecast_id,
        obs.observation_id, ref_forecast_id, fxagg.forecast_id,
        agg.aggregate_id,
    ]
    mocked = requests_mock.register_uri(
        'POST', re.compile(f'{session.base_url}/reports/.*/values'),
        [{'text': id_} for id_ in ids])
    inp = raw_report(True)
    out = session.post_raw_report_processed_data('report_id', inp)
    exp = raw_report(False)
    assert out == exp.processed_forecasts_observations
    history = mocked.request_history
    for i, id_ in enumerate(ids):
        assert id_ == history[i].json()['object_id']


def test_apisession_get_raw_report_processed_data(
        requests_mock, raw_report, report_objects, ref_forecast_id):
    _, obs, fx0, fx1, agg, fxagg = report_objects
    session = api.APISession('')
    ser = pd.Series(name='value', index=pd.DatetimeIndex(
        [], tz='UTC', name='timestamp'), dtype=float)
    val = utils.serialize_timeseries(ser)
    requests_mock.register_uri(
        'GET', re.compile(f'{session.base_url}/reports/.*/values'),
        json=[{'id': id_, 'processed_values': val} for id_ in
              (fx0.forecast_id, fx1.forecast_id, obs.observation_id,
               ref_forecast_id, agg.aggregate_id, fxagg.forecast_id)])
    inp = raw_report(False)
    out = session.get_raw_report_processed_data('', inp)
    for fxo in out:
        pdt.assert_series_equal(fxo.forecast_values, ser)
        pdt.assert_series_equal(fxo.observation_values, ser)
        if fxo.reference_forecast_values is not None:
            pdt.assert_series_equal(fxo.reference_forecast_values, ser)


def test_apisession_post_raw_report(requests_mock, raw_report, mocker,
                                    report_objects):
    raw = raw_report(True)
    _, obs, fx0, fx1, agg, fxagg = report_objects
    session = api.APISession('')
    ids = [fx0.forecast_id, obs.observation_id, fx1.forecast_id,
           obs.observation_id, agg.aggregate_id, fxagg.forecast_id]
    requests_mock.register_uri(
        'POST', re.compile(f'{session.base_url}/reports/.*/values'),
        [{'text': id_} for id_ in ids])
    mocked = requests_mock.register_uri(
        'POST', re.compile(f'{session.base_url}/reports/.*/raw'))
    status = mocker.patch(
        'solarforecastarbiter.io.api.APISession.update_report_status')
    session.post_raw_report('', raw)
    assert isinstance(mocked.last_request.json(), dict)
    assert status.called


def test_apisession_update_report_status(requests_mock):
    session = api.APISession('')
    mocked = requests_mock.register_uri(
        'POST', re.compile(f'{session.base_url}/reports/REPORT_ID/status/'))
    session.update_report_status('REPORT_ID', 'complete')
    assert mocked.last_request.url.split('/')[-1] == 'complete'


@pytest.fixture()
def mockobs(many_observations, mocker):
    def _getobs(cls, observation_id):
        for obs in many_observations:
            if observation_id == obs.observation_id:
                return obs

    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_observation',
        new=_getobs)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.list_observations',
        return_value=many_observations)


def test_apisession_get_aggregate(requests_mock, aggregate_text, aggregate,
                                  mockobs):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET', re.compile(f'{session.base_url}/aggregates/.*/metadata'),
        content=aggregate_text)
    agg = session.get_aggregate('')
    assert agg == aggregate


def test_apisession_list_aggregates(requests_mock, aggregate_text, aggregate,
                                    mockobs):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET', re.compile(f'{session.base_url}/aggregates/'),
        content=b'[' + aggregate_text + b']')
    aggs = session.list_aggregates()
    assert aggs[0] == aggregate


def test_apisession_list_aggregates_single(requests_mock, aggregate_text,
                                           aggregate, mockobs):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET', re.compile(f'{session.base_url}/aggregates/'),
        content=aggregate_text)
    aggs = session.list_aggregates()
    assert aggs[0] == aggregate


def test_apisession_list_aggregates_empty(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/aggregates/.*')
    requests_mock.register_uri('GET', matcher, content=b"[]")
    obs_list = session.list_aggregates()
    assert obs_list == []


def test_apisession_create_aggregate(requests_mock, aggregate, aggregate_text,
                                     mockobs):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/aggregates/.*')

    def callback(request, context):
        if request.url.endswith('aggregates/'):
            ad = json.loads(aggregate_text)
            for key in ('provider', 'aggregate_id', 'created_at',
                        'modified_at', 'observations', 'interval_value_type'):
                del ad[key]
            assert ad == request.json()
            return aggregate.aggregate_id
        else:
            rj = request.json()
            assert 'observations' in rj
            assert 'observation_id' in rj['observations'][0]
            assert (
                'effective_from' in rj['observations'][0]
                or 'effective_until' in rj['observations'][0])

    requests_mock.register_uri('POST', matcher, text=callback)
    requests_mock.register_uri('GET', matcher, content=aggregate_text)
    aggregate_dict = aggregate.to_dict()
    del aggregate_dict['aggregate_id']
    del aggregate_dict['interval_value_type']
    ss = datamodel.Aggregate.from_dict(aggregate_dict)
    new_aggregate = session.create_aggregate(ss)
    assert new_aggregate == aggregate


def test_apisession_get_user_info(requests_mock):
    session = api.APISession('')
    matcher = re.compile(f'{session.base_url}/users/current')
    requests_mock.register_uri('GET', matcher, content=b'{"test": "key"}')
    out = session.get_user_info()
    assert out['test'] == 'key'


time_range_params = pytest.mark.parametrize('mint,maxt,expected', [
    ('"2020-03-03T11:34:23+00:00"', '"2020-03-04T13:00:00+00:00"',
     (pd.Timestamp(year=2020, month=3, day=3, hour=11, minute=34, second=23,
                   tz='UTC'),
      pd.Timestamp(year=2020, month=3, day=4, hour=13, tz='UTC'))),
    ('null', 'null', (pd.NaT, pd.NaT)),
    ('"2020-03-03T11:34:23"', '"2020-03-04T13:00:00+00:00"',
     (pd.Timestamp(year=2020, month=3, day=3, hour=11, minute=34, second=23,
                   tz='UTC'),
      pd.Timestamp(year=2020, month=3, day=4, hour=13, tz='UTC'))),
    ('"2020-03-03T11:00:00+00:00"', '"2020-03-11T23:01:00"',
     (pd.Timestamp(year=2020, month=3, day=3, hour=11, tz='UTC'),
      pd.Timestamp(year=2020, month=3, day=11, hour=23, minute=1,
                   tz='UTC')))
])


@time_range_params
def test_get_observation_time_range(requests_mock, mint, maxt, expected):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET', f'{session.base_url}/observations/obsid/values/timerange',
        content=(
            '{"observation_id": "obsid", "min_timestamp": '
            f'{mint}, "max_timestamp": {maxt}' + '}').encode())
    out = session.get_observation_time_range('obsid')
    assert out == expected


@time_range_params
def test_get_cdf_forecast_time_range(requests_mock, mint, maxt, expected):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET',
        f'{session.base_url}/forecasts/cdf/single/fxid/values/timerange',
        content=(
            '{"forecast_id": "fxid", "min_timestamp": '
            f'{mint}, "max_timestamp": {maxt}' + '}').encode())
    out = session.get_probabilistic_forecast_constant_value_time_range('fxid')
    assert out == expected


@time_range_params
def test_get_forecast_time_range(requests_mock, mint, maxt, expected):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET', f'{session.base_url}/forecasts/single/fxid/values/timerange',
        content=(
            '{"forecast_id": "fxid", "min_timestamp": '
            f'{mint}, "max_timestamp": {maxt}' + '}').encode())
    out = session.get_forecast_time_range('fxid')
    assert out == expected


@pytest.mark.parametrize('inp,exp', [
    ('[]', np.array([], dtype='datetime64[D]')),
    ('["2019-01-01"]',
     np.array(['2019-01-01'], dtype='datetime64[D]')),
    ('["2019-01-03", "2019-04-01"]',
     np.array(['2019-01-03', '2019-04-01'],
              dtype='datetime64[D]')),
])
def test_get_observation_values_not_flagged(requests_mock, inp, exp):
    session = api.APISession('')
    requests_mock.register_uri(
        'GET', f'{session.base_url}/observations/obsid/values/unflagged',
        content=(
            '{"observation_id": "obsid", "_links": {}, '
            f'"dates": {inp}' + '}').encode())
    out = session.get_observation_values_not_flagged(
        'obsid', '2019-01-01T00:00Z', '2020-01-01T00:00Z', 16)
    assert all(out == exp)


@pytest.fixture(scope='session')
def auth_token():
    try:
        token_req = requests.post(
            'https://solarforecastarbiter.auth0.com/oauth/token',
            headers={'content-type': 'application/json'},
            data=('{"grant_type": "password", '
                  '"username": "testing@solarforecastarbiter.org",'
                  '"password": "Thepassword123!", '
                  '"audience": "https://api.solarforecastarbiter.org", '
                  '"client_id": "c16EJo48lbTCQEhqSztGGlmxxxmZ4zX7"}'))
    except Exception:
        return pytest.skip('Cannot retrieve valid Auth0 token')
    else:
        token = token_req.json()['access_token']
        return token


@pytest.fixture(scope='session')
def real_session(auth_token):
    session = api.APISession(
        auth_token, base_url='https://dev-api.solarforecastarbiter.org')
    try:
        session.get('')
    except Exception:
        return pytest.skip('Cannot connect to dev api')
    else:
        return session


def test_real_apisession_get_site(real_session):
    site = real_session.get_site('123e4567-e89b-12d3-a456-426655440001')
    assert isinstance(site, datamodel.Site)


def test_real_apisession_list_sites(real_session):
    sites = real_session.list_sites()
    assert isinstance(sites, list)
    assert isinstance(sites[0], datamodel.Site)


def test_real_apisession_list_sites_in_zone(real_session):
    sites = real_session.list_sites_in_zone('Reference Region 5')
    assert isinstance(sites, list)
    assert isinstance(sites[0], datamodel.Site)


@pytest.mark.parametrize('lat,lon,expected', [
    (32.1, -110.8, {'Reference Region 3'}),
    (45, -70.0, {'Reference Region 7'}),
    (0, 0, set())
])
def test_real_apisession_search_climatezones(real_session, lat, lon, expected):
    zones = real_session.search_climatezones(lat, lon)
    assert set(zones) == expected


def test_real_apisession_create_site(site_text, real_session):
    sited = json.loads(site_text)
    sited['name'] = f'Test create site {randint(0, 100)}'
    site = datamodel.SolarPowerPlant.from_dict(sited)
    new_site = real_session.create_site(site)
    real_session.delete(f'/sites/{new_site.site_id}')
    assert new_site.name == site.name
    assert new_site.modeling_parameters == site.modeling_parameters


def test_real_apisession_get_observation(real_session):
    obs = real_session.get_observation('123e4567-e89b-12d3-a456-426655440000')
    assert isinstance(obs, datamodel.Observation)


def test_real_apisession_list_observations(real_session):
    obs = real_session.list_observations()
    assert isinstance(obs, list)
    assert isinstance(obs[0], datamodel.Observation)


def test_real_apisession_create_observation(single_observation_text,
                                            single_observation,
                                            real_session):
    observationd = json.loads(single_observation_text)
    observationd['name'] = f'Test create observation {randint(0, 100)}'
    observationd['site'] = single_observation.site
    observation = datamodel.Observation.from_dict(observationd)
    new_observation = real_session.create_observation(observation)
    real_session.delete(f'/observations/{new_observation.observation_id}')
    for attr in ('name', 'site', 'variable', 'interval_label',
                 'interval_length', 'extra_parameters', 'uncertainty'):
        assert getattr(new_observation, attr) == getattr(observation, attr)


def test_real_apisession_get_forecast(real_session):
    fx = real_session.get_forecast('f8dd49fa-23e2-48a0-862b-ba0af6dec276')
    assert isinstance(fx, datamodel.Forecast)


def test_real_apisession_list_forecasts(real_session):
    fxs = real_session.list_forecasts()
    assert isinstance(fxs, list)
    assert isinstance(fxs[0], datamodel.Forecast)


def test_real_apisession_create_forecast(single_forecast_text, single_forecast,
                                         real_session):
    forecastd = json.loads(single_forecast_text)
    forecastd['name'] = f'Test create forecast {randint(0, 100)}'
    forecastd['site'] = single_forecast.site
    forecast = datamodel.Forecast.from_dict(forecastd)
    new_forecast = real_session.create_forecast(forecast)
    real_session.delete(f'/forecasts/single/{new_forecast.forecast_id}')
    for attr in ('name', 'site', 'variable', 'interval_label',
                 'interval_length', 'lead_time_to_start', 'run_length',
                 'extra_parameters'):
        assert getattr(new_forecast, attr) == getattr(forecast, attr)


def test_real_apisession_create_forecast_invalid(
        single_forecast_text, single_forecast, real_session):
    forecastd = json.loads(single_forecast_text)
    forecastd['name'] = f'Test create forecast {randint(0, 100)}'
    forecastd['site'] = single_forecast.site
    forecastd['interval_label'] = 'mean'
    forecast = datamodel.Forecast.from_dict(forecastd)
    with pytest.raises(requests.exceptions.HTTPError) as e:
        real_session.create_forecast(forecast)
    assert 'Must be one of' in str(e.value)


def test_real_apisession_get_prob_forecast(real_session):
    fx = real_session.get_probabilistic_forecast(
        'ef51e87c-50b9-11e9-8647-d663bd873d93')
    assert isinstance(fx, datamodel.ProbabilisticForecast)


def test_real_apisession_list_prob_forecasts(real_session):
    fxs = real_session.list_probabilistic_forecasts()
    assert isinstance(fxs, list)
    assert isinstance(fxs[0], datamodel.ProbabilisticForecast)


def test_real_apisession_create_prob_forecast(
        prob_forecast_text, prob_forecasts, prob_forecast_constant_value,
        real_session):
    forecastd = json.loads(prob_forecast_text)
    forecastd['name'] = f'Test create forecast {randint(0, 100)}'
    forecastd['site'] = prob_forecasts.site
    fx_prob_cv_d = prob_forecast_constant_value.to_dict()
    fx_prob_cv_d['name'] = forecastd['name']
    del fx_prob_cv_d['forecast_id']
    forecastd['constant_values'] = (
        datamodel.ProbabilisticForecastConstantValue.from_dict(fx_prob_cv_d), )
    forecast = datamodel.ProbabilisticForecast.from_dict(forecastd)
    new_forecast = real_session.create_probabilistic_forecast(forecast)
    real_session.delete(f'/forecasts/cdf/{new_forecast.forecast_id}')
    test_attrs = ('name', 'site', 'variable', 'interval_label',
                  'interval_length', 'lead_time_to_start', 'run_length',
                  'extra_parameters')
    for attr in test_attrs:
        assert getattr(new_forecast, attr) == getattr(forecast, attr)
    for new_cv, cv in zip(new_forecast.constant_values,
                          forecast.constant_values):
        for attr in test_attrs:
            assert getattr(new_cv, attr) == getattr(cv, attr)


def test_real_apisession_get_prob_forecast_constant_value(real_session):
    fx = real_session.get_probabilistic_forecast_constant_value(
        '633f9b2a-50bb-11e9-8647-d663bd873d93')
    assert isinstance(fx, datamodel.ProbabilisticForecastConstantValue)


def test_real_apisession_get_observation_values(real_session):
    start = pd.Timestamp('2019-04-15T00:00:00Z')
    end = pd.Timestamp('2019-04-15T12:00:00Z')
    obs = real_session.get_observation_values(
        '123e4567-e89b-12d3-a456-426655440000',
        start, end)
    assert isinstance(obs, pd.DataFrame)
    assert set(obs.columns) == {'value', 'quality_flag'}
    assert len(obs.index) > 0
    pdt.assert_frame_equal(obs.loc[start:end], obs)


def test_real_apisession_get_observation_values_tz(real_session):
    # use different tzs to confirm that it works
    start = pd.Timestamp('2019-04-14T22:00:00-0200')
    end = pd.Timestamp('2019-04-15T05:00:00-0700')
    obs = real_session.get_observation_values(
        '123e4567-e89b-12d3-a456-426655440000',
        start, end)
    assert isinstance(obs, pd.DataFrame)
    assert set(obs.columns) == {'value', 'quality_flag'}
    assert len(obs.index) > 0
    end = end.tz_convert(start.tzinfo)
    pdt.assert_frame_equal(obs.loc[start:end], obs)


def test_real_apisession_get_forecast_values(real_session):
    start = pd.Timestamp('2019-04-15T00:00:00Z')
    end = pd.Timestamp('2019-04-15T12:00:00Z')
    fx = real_session.get_forecast_values(
        'f8dd49fa-23e2-48a0-862b-ba0af6dec276',
        start, end)
    assert isinstance(fx, pd.Series)
    assert len(fx) > 0
    pdt.assert_series_equal(fx.loc[start:end], fx)


def test_real_apisession_get_forecast_values_tz(real_session):
    # use different tzs to confirm that it works
    start = pd.Timestamp('2019-04-14T20:00:00-0400')
    end = pd.Timestamp('2019-04-15T13:00:00+0100')
    fx = real_session.get_forecast_values(
        'f8dd49fa-23e2-48a0-862b-ba0af6dec276',
        start, end)
    assert isinstance(fx, pd.Series)
    assert len(fx) > 0
    end = end.tz_convert(start.tzinfo)
    pdt.assert_series_equal(fx.loc[start:end], fx)


def test_real_apisession_get_prob_forecast_values_tz(real_session):
    # use different tzs to confirm that it works
    start = pd.Timestamp('2019-04-14T20:00:00-0400')
    end = pd.Timestamp('2019-04-15T13:00:00+0100')
    fx = real_session.get_probabilistic_forecast_constant_value_values(
        '633f9b2a-50bb-11e9-8647-d663bd873d93',
        start, end)
    assert isinstance(fx, pd.Series)
    assert len(fx) > 0
    end = end.tz_convert(start.tzinfo)
    pdt.assert_series_equal(fx.loc[start:end], fx)


def test_real_apisession_post_observation_values(real_session):
    test_df = pd.DataFrame(
        {'value': [np.random.random()], 'quality_flag': [0]},
        index=pd.DatetimeIndex([pd.Timestamp('2019-04-14T00:00:00Z')],
                               name='timestamp'))
    real_session.post_observation_values(
        '123e4567-e89b-12d3-a456-426655440000', test_df)
    obs = real_session.get_observation_values(
        '123e4567-e89b-12d3-a456-426655440000',
        pd.Timestamp('2019-04-14T00:00:00Z'),
        pd.Timestamp('2019-04-14T00:01:00Z'))
    # quality flag may be altered by validation routine
    pdt.assert_series_equal(obs['value'], test_df['value'])


def test_real_apisession_post_forecast_values(real_session):
    test_ser = pd.Series(
        [np.random.random()], name='value',
        index=pd.DatetimeIndex([pd.Timestamp('2019-04-14T00:00:00Z')],
                               name='timestamp'))
    real_session.post_forecast_values(
        'f8dd49fa-23e2-48a0-862b-ba0af6dec276', test_ser)
    fx = real_session.get_forecast_values(
        'f8dd49fa-23e2-48a0-862b-ba0af6dec276',
        pd.Timestamp('2019-04-14T00:00:00Z'),
        pd.Timestamp('2019-04-14T00:01:00Z'))
    pdt.assert_series_equal(fx, test_ser)


def test_real_apisession_post_prob_forecast_constant_val_values(real_session):
    test_ser = pd.Series(
        [np.random.random()], name='value',
        index=pd.DatetimeIndex([pd.Timestamp('2019-04-14T00:00:00Z')],
                               name='timestamp'))
    real_session.post_probabilistic_forecast_constant_value_values(
        '633f9b2a-50bb-11e9-8647-d663bd873d93', test_ser)
    fx = real_session.get_probabilistic_forecast_constant_value_values(
        '633f9b2a-50bb-11e9-8647-d663bd873d93',
        pd.Timestamp('2019-04-14T00:00:00Z'),
        pd.Timestamp('2019-04-14T00:01:00Z'))
    pdt.assert_series_equal(fx, test_ser)


def test_real_apisession_get_aggregate(real_session):
    agg = real_session.get_aggregate('458ffc27-df0b-11e9-b622-62adb5fd6af0')
    assert isinstance(agg, datamodel.Aggregate)


def test_real_apisession_list_aggregates(real_session):
    aggs = real_session.list_aggregates()
    assert isinstance(aggs, list)
    assert isinstance(aggs[0], datamodel.Aggregate)
    assert '458ffc27-df0b-11e9-b622-62adb5fd6af0' in {
        a.aggregate_id for a in aggs}


def test_real_apisession_create_aggregate(real_session, aggregate):
    new_agg = real_session.create_aggregate(aggregate)
    real_session.delete(f'/aggregates/{new_agg.aggregate_id}')
    for attr in ('name', 'description', 'variable', 'aggregate_type',
                 'interval_length', 'interval_label', 'timezone'):
        assert getattr(new_agg, attr) == getattr(aggregate, attr)
    for i, obs in enumerate(new_agg.observations):
        assert (
            aggregate.observations[i].observation.observation_id ==
            obs.observation.observation_id)
        for attr in ('effective_from', 'effective_until',
                     'observation_deleted_at'):
            assert getattr(aggregate.observations[i], attr) == (
                getattr(obs, attr))


def test_real_apisession_get_aggregate_values(real_session):
    start = pd.Timestamp('2019-04-15T00:00:00Z')
    end = pd.Timestamp('2019-04-15T12:00:00Z')
    agg = real_session.get_aggregate_values(
        '458ffc27-df0b-11e9-b622-62adb5fd6af0',
        start, end)
    assert isinstance(agg, pd.DataFrame)
    assert set(agg.columns) == {'value', 'quality_flag'}
    assert len(agg.index) > 0
    pdt.assert_frame_equal(agg.loc[start:end], agg)


def test_real_apisession_get_user_info(real_session):
    user_info = real_session.get_user_info()
    assert user_info['organization'] == 'Organization 1'


def test_real_apisession_get_observation_time_range(real_session):
    out = real_session.get_observation_time_range(
        '123e4567-e89b-12d3-a456-426655440000')
    assert out == (
        pd.Timestamp('2019-04-14T00:00:00Z'),
        pd.Timestamp('2019-04-17T06:55:00Z'))


def test_real_apisession_get_forecast_time_range(real_session):
    out = real_session.get_forecast_time_range(
        'f8dd49fa-23e2-48a0-862b-ba0af6dec276')
    assert out == (
        pd.Timestamp('2019-04-14T00:00:00Z'),
        pd.Timestamp('2019-04-17T06:59:00Z'))


def test_real_apisession_get_cdf_forecast_time_range(real_session):
    out = real_session.get_probabilistic_forecast_constant_value_time_range(
        '633f9b2a-50bb-11e9-8647-d663bd873d93')
    assert out == (
        pd.Timestamp('2019-04-14T00:00:00Z'),
        pd.Timestamp('2019-04-17T06:55:00Z'))


def test_real_apisession_get_observation_values_not_flagged(real_session):
    start = pd.Timestamp('2019-04-15T00:00:00Z')
    end = pd.Timestamp('2019-04-15T12:00:00Z')
    out = real_session.get_observation_values_not_flagged(
        '123e4567-e89b-12d3-a456-426655440000',
        start, end, 1)
    assert isinstance(out, np.ndarray)
    assert out.dtype == 'datetime64[D]'
    assert all(out == np.array(['2019-04-15'], dtype='datetime64[D]'))


@pytest.mark.parametrize('ftype,expected_fn', [
    ('forecast', 'get_forecast'),
    ('event_forecast', 'get_forecast'),
    ('probabilistic_forecast', 'get_probabilistic_forecast'),
    ('probabilistic_forecast_constant_value',
     'get_probabilistic_forecast_constant_value'),
])
def test_api_session_forecast_get_by_type(ftype, expected_fn):
    test_session = api.APISession('token')
    get_fn = test_session._forecast_get_by_type(ftype)
    assert get_fn == getattr(test_session, expected_fn)


def test_api_session_forecast_get_by_type_invalid_type():
    test_session = api.APISession('token')
    with pytest.raises(ValueError):
        test_session._forecast_get_by_type('bad_type')



import itertools

import pandas as pd
from pandas.testing import assert_series_equal

import pytest

from solarforecastarbiter.reference_forecasts import forecast


def assert_none_or_series(out, expected):
    assert len(out) == len(expected)
    for o, e in zip(out, expected):
        if e is None:
            assert o is None
        else:
            assert_series_equal(o, e)


def test_resample():
    index = pd.date_range(start='20190101', freq='15min', periods=5)
    arg = pd.Series([1, 0, 0, 0, 2], index=index)
    idx_exp = pd.date_range(start='20190101', freq='1h', periods=2)
    expected = pd.Series([0.25, 2.], index=idx_exp)
    out = forecast.resample(arg)
    assert_series_equal(out, expected)
    assert forecast.resample(None) is None


@pytest.fixture
def rfs_series():
    return pd.Series([1, 2],
                     index=pd.DatetimeIndex(['20190101 01', '20190101 02']))


@pytest.mark.parametrize(
    'start,end,start_slice,end_slice,fill_method,exp_val,exp_idx', [
        (None, None, None, None, 'interpolate', [1, 1.5, 2],
         ['20190101 01', '20190101 0130', '20190101 02']),
        ('20190101', '20190101 0230', None, None, 'interpolate',
         [1, 1, 1, 1.5, 2, 2],
         ['20190101', '20190101 0030', '20190101 01', '20190101 0130',
          '20190101 02', '20190101 0230']),
        ('20190101', '20190101 02', '20190101 0030', '20190101 0130', 'bfill',
         [1., 1, 2], ['20190101 0030', '20190101 01', '20190101 0130'])
    ]
)
def test_reindex_fill_slice(rfs_series, start, end, start_slice, end_slice,
                            fill_method, exp_val, exp_idx):
    exp = pd.Series(exp_val, index=pd.DatetimeIndex(exp_idx))
    out = forecast.reindex_fill_slice(
        rfs_series, freq='30min', start=start, end=end,
        start_slice=start_slice, end_slice=end_slice, fill_method=fill_method)
    assert_series_equal(out, exp)


def test_reindex_fill_slice_some_nan():
    rfs_series = pd.Series([1, 2, None, 4], index=pd.DatetimeIndex([
        '20190101 01', '20190101 02', '20190101 03', '20190101 04',
    ]))
    start, end, start_slice, end_slice, fill_method = \
        None, None, None, None, 'interpolate'
    exp_val = [1, 1.5, 2, 2.5, 3, 3.5, 4]
    exp_idx = [
        '20190101 01', '20190101 0130', '20190101 02', '20190101 0230',
        '20190101 03', '20190101 0330', '20190101 04']
    exp = pd.Series(exp_val, index=pd.DatetimeIndex(exp_idx))
    out = forecast.reindex_fill_slice(
        rfs_series, freq='30min', start=start, end=end,
        start_slice=start_slice, end_slice=end_slice, fill_method=fill_method)
    assert_series_equal(out, exp)


def test_reindex_fill_slice_all_nan():
    arg = pd.Series([None]*3, index=pd.DatetimeIndex(
        ['20190101 01', '20190101 02', '20190101 03']))
    out = forecast.reindex_fill_slice(arg, freq='30min')
    exp = pd.Series([None]*5, index=pd.DatetimeIndex(
        ['20190101 01', '20190101 0130', '20190101 02', '20190101 0230',
         '20190101 03']))
    assert_series_equal(out, exp)


def test_reindex_fill_slice_empty():
    out = forecast.reindex_fill_slice(pd.Series(dtype=float), freq='30min')
    assert_series_equal(out, pd.Series(dtype=float))


def test_reindex_fill_slice_none():
    out = forecast.reindex_fill_slice(None, freq='30min')
    assert out is None


def test_cloud_cover_to_ghi_linear():
    cloud_cover = pd.Series([0, 50, 100.])
    ghi_clear = pd.Series([1000, 1000, 1000.])
    out = forecast.cloud_cover_to_ghi_linear(cloud_cover, ghi_clear)
    expected = pd.Series([1000, 675, 350.])
    assert_series_equal(out, expected)
    out = forecast.cloud_cover_to_ghi_linear(cloud_cover, ghi_clear, offset=20)
    expected = pd.Series([1000, 600, 200.])
    assert_series_equal(out, expected)


@pytest.mark.xfail(raises=AssertionError, strict=True)
def test_cloud_cover_to_irradiance_ghi_clear():
    index = pd.date_range(start='20190101', periods=3, freq='1h')
    cloud_cover = pd.Series([0, 50, 100.], index=index)
    ghi_clear = pd.Series([10, 10, 1000.], index=index)
    zenith = pd.Series([90.0, 89.9, 45], index=index)
    out = forecast.cloud_cover_to_irradiance_ghi_clear(
        cloud_cover, ghi_clear, zenith
    )
    # https://github.com/pvlib/pvlib-python/issues/681
    ghi_exp = pd.Series([10., 6.75, 350.])
    dni_exp = pd.Series([0., 0., 4.74198165e+01])
    dhi_exp = pd.Series([10., 6.75, 316.46912616])
    assert_series_equal(out[0], ghi_exp)
    assert_series_equal(out[1], dni_exp)
    assert_series_equal(out[2], dhi_exp)


@pytest.mark.xfail(raises=AssertionError, strict=True)
def test_cloud_cover_to_irradiance():
    index = pd.date_range(start='20190101', periods=3, freq='1h')
    cloud_cover = pd.Series([0, 50, 100.], index=index)
    latitude = 32.2
    longitude = -110.9
    elevation = 700
    zenith = pd.Series([90.0, 89.9, 45], index=index)
    apparent_zenith = pd.Series([89.9, 89.85, 45], index=index)
    out = forecast.cloud_cover_to_irradiance(
        latitude, longitude, elevation, cloud_cover, apparent_zenith, zenith
    )
    # https://github.com/pvlib/pvlib-python/issues/681
    ghi_exp = pd.Series([10., 6.75, 350.], index=index)
    dni_exp = pd.Series([0., 0., 4.74198165e+01], index=index)
    dhi_exp = pd.Series([10., 6.75, 316.46912616], index=index)
    assert_series_equal(out[0], ghi_exp)
    assert_series_equal(out[1], dni_exp)
    assert_series_equal(out[2], dhi_exp)


@pytest.mark.parametrize('mixed,expected', [
    ([1, 1/2, 1/3, 1/4, 1/5, 1/6], [1., 0, 0, 0, 0, 0]),
    ([0, 0, 0, 0, 0, 1/6], [0, 0, 0, 0, 0, 1.]),
    ([0, 0, 0, 0, 0, 1/6, 1, 1/2, 1/3, 1/4, 1/5, 1/6],
     [0, 0, 0, 0, 0, 1., 1., 0, 0, 0, 0, 0]),
    ([65.0, 66.0, 44.0, 32.0, 30.0, 26.0],  # GH 144
     [65.0, 67.0, 0.0, 0.0, 22.0, 6.0]),  # 4th element is -4 if no clipping
    ([1, 1/2], [1., 0]),
    ([0, 1/2], [0, 1.]),
    ([0, 1/2, 1, 1/2], [0, 1., 1., 0])
])
def test_unmix_intervals(mixed, expected):
    npts = len(mixed)
    if npts in [2, 4]:
        index = pd.date_range(start='20190101 03Z', freq='3h', periods=npts)
    else:
        index = pd.date_range(start='20190101 01Z', freq='1h', periods=npts)
    mixed_s = pd.Series(mixed, index=index)
    out = forecast.unmix_intervals(mixed_s)
    expected_s = pd.Series(expected, index=index)
    assert_series_equal(out, expected_s)


# allowed times are 1, 7, 13, 19Z
_1h_allowed_0700 = [0, 6, 12, 18]
_1h_not_allowed_0700 = [x for x in range(0, 24) if x not in _1h_allowed_0700]
# allowed times are 3, 9, 15, 21Z
_3h_allowed_0700 = [2, 8, 14, 20]
_3h_not_allowed_0700 = [x for x in range(0, 24) if x not in _3h_allowed_0700]


@pytest.mark.parametrize('hr,freq', (
    list(itertools.product(_1h_allowed_0700, ['1h'])) +
    list(itertools.product(_3h_allowed_0700, ['3h']))
))
def test_unmix_intervals_tz(hr, freq):
    periods = 6 if freq == '1h' else 2
    # should work.
    index = pd.date_range(
        start=f'20190101 {hr:02}-0700', freq=freq, periods=periods)
    # we only care to test tz functionality, not values
    mixed_s = pd.Series([0] * len(index), index=index)
    out = forecast.unmix_intervals(mixed_s)
    expected_s = pd.Series([0] * len(index), index=index)
    assert_series_equal(out, expected_s)


@pytest.mark.parametrize('hr,freq', (
    list(itertools.product(_1h_not_allowed_0700, ['1h'])) +
    list(itertools.product(_3h_not_allowed_0700, ['3h']))
))
def test_unmix_intervals_tz_fail(hr, freq):
    periods = 6 if freq == '1h' else 2
    # should not work. 13-0700 is 20Z
    index = pd.date_range(
        start=f'20190101 {hr:02}-0700', freq=freq, periods=periods)
    with pytest.raises(ValueError):
        mixed_s = pd.Series([0] * len(index), index=index)
        forecast.unmix_intervals(mixed_s)


def test_unmix_intervals_invalidfreq():
    index = pd.date_range(start='20190101 01Z', freq='2h', periods=3)
    mixed_s = pd.Series([1, 1/3, 1/6], index=index)
    with pytest.raises(ValueError):
        forecast.unmix_intervals(mixed_s)


def test_unmix_intervals_two_freq():
    index = pd.DatetimeIndex(['20190101 01', '20190101 02', '20190101 04'])
    mixed_s = pd.Series([1, 1/3, 1/6], index=index)
    with pytest.raises(ValueError):
        forecast.unmix_intervals(mixed_s)



import asyncio
import atexit
from functools import partial, wraps
import logging
import multiprocessing as mp
import signal
import threading


import aiohttp


CLUSTER = None


def ignore_interrupt():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def start_cluster(max_workers=4, maxtasksperchild=5):
    global CLUSTER
    mp.set_start_method("forkserver")
    CLUSTER = mp.Pool(max_workers, maxtasksperchild=maxtasksperchild,
                      initializer=ignore_interrupt)
    atexit.register(CLUSTER.terminate)
    return


async def run_in_executor(func, *args, **kwargs):
    exc = partial(CLUSTER.apply, func, args, kwargs)
    # uses the asyncio default thread pool executor to then
    # apply the function on the pool of processes
    # inefficient, but ProcessPoolExecutor will not restart
    # processes in case of memory leak
    res = await asyncio.get_event_loop().run_in_executor(None, exc)
    return res


def make_session():
    """Make an aiohttp session"""
    conn = aiohttp.TCPConnector(limit_per_host=20)
    timeout = aiohttp.ClientTimeout(total=60, connect=10, sock_read=30)
    s = aiohttp.ClientSession(connector=conn, timeout=timeout)
    return s


def abort_all_on_exception(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            ret = await f(*args, **kwargs)
        except Exception:
            logging.exception('Aborting on error')
            signal.pthread_kill(threading.get_ident(), signal.SIGUSR1)
        else:
            return ret
    return wrapper



"""
Functions to make all of the figures for Solar Forecast Arbiter reports using
Bokeh.

This code is currently unreachable from the rest of the Solar Forecast Arbiter
Core library. It may be used in place of the plotly_figures to generate bokeh
plots for the `plots` attribute of the RawReport object. See
:py:mod:`solarforecastarbiter.reports.main` for an example of report
generation.
"""
import calendar
from contextlib import contextmanager
import datetime as dt
from itertools import cycle
import logging
import warnings


from bokeh.embed import components
from bokeh.io.export import get_svgs
from bokeh.layouts import gridplot
from bokeh.models import (ColumnDataSource, HoverTool, Legend,
                          DatetimeTickFormatter, CategoricalTickFormatter,
                          CDSView, GroupFilter, BooleanFilter)
from bokeh.models.ranges import Range1d, FactorRange, DataRange1d
from bokeh.plotting import figure
from bokeh.transform import factor_cmap, dodge
from bokeh import palettes
from bokeh import __version__ as bokeh_version
import pandas as pd
import numpy as np


from solarforecastarbiter import datamodel
from solarforecastarbiter.plotting.utils import line_or_step


logger = logging.getLogger(__name__)
PALETTE = (
    palettes.d3['Category20'][20][::2] + palettes.d3['Category20'][20][1::2])
_num_obs_colors = 3
# drop white
OBS_PALETTE = list(palettes.grey(_num_obs_colors + 1)[0:_num_obs_colors])
OBS_PALETTE.reverse()
OBS_PALETTE_TD_RANGE = pd.timedelta_range(
    freq='10min', end='60min', periods=_num_obs_colors)


def construct_timeseries_cds(report):
    """Construct two standardized Bokeh CDS for the timeseries and scatter
    plot functions. One with timeseries data for all observations,
    aggregates, and forecasts in the report, and the other with
    associated metadata sharing a common `pair_index` key.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`

    Returns
    -------
    value_cds : bokeh.models.ColumnDataSource
        Keys are an integer `pair_index` for pairing values with the metadata
        in the metadata_cds, and two pandas.Series, `observation_values` and
        `forecast_values`.

    metadata_cds : bokeh.models.ColumnDataSource
        This cds has the following keys:

        - `pair_index`: Integer for pairing metadata with the values in the value_cds.
        - `observation_name`: Observation name.
        - `forecast_name`: Forecast name.
        - `interval_label`: Interval label of the processed forecast and observation data.
        - `observation_hash`: Hash of the original observation object and the `datamodel.ProcessedForecastObservations` metadata.
        - `forecast_hash`: Hash of the original forecast object and the `datamodel.ProcessedForecastObservations` metadata.

    """  # NOQA
    value_frames = []
    meta_rows = []
    for idx, pfxobs in enumerate(
            report.raw_report.processed_forecasts_observations):
        value_frame_dict = {
            'pair_index': idx,
            'observation_values': pfxobs.observation_values,
            'forecast_values': pfxobs.forecast_values,
        }
        meta_row_dict = {
            'pair_index': idx,
            'observation_name': _obs_name(pfxobs.original),
            'forecast_name': _fx_name(pfxobs.original),
            'interval_label': pfxobs.interval_label,
            'observation_hash': str(hash(
                (pfxobs.original.data_object,
                 pfxobs.interval_length,
                 pfxobs.interval_value_type,
                 pfxobs.interval_label))),
            'forecast_hash': str(hash(
                (pfxobs.original.forecast,
                 pfxobs.interval_length,
                 pfxobs.interval_value_type,
                 pfxobs.interval_label))),
            'observation_color': _obs_color(
                pfxobs.interval_length)
        }
        value_frames.append(pd.DataFrame(value_frame_dict))
        meta_rows.append(meta_row_dict)
    data = pd.concat(value_frames)
    metadata = pd.DataFrame(meta_rows)
    # drop tz info from localized times. GH164
    data = data.tz_localize(None)
    data = data.rename_axis('timestamp')
    value_cds = ColumnDataSource(data)
    metadata_cds = ColumnDataSource(metadata)
    return value_cds, metadata_cds


def _obs_name(fx_obs):
    # TODO: add code to ensure obs names are unique
    name = fx_obs.data_object.name
    if fx_obs.forecast.name == fx_obs.data_object.name:
        if isinstance(fx_obs.data_object, datamodel.Observation):
            name += ' Observation'
        else:
            name += ' Aggregate'
    return name


def _fx_name(fx_obs):
    # TODO: add code to ensure fx names are unique
    name = fx_obs.forecast.name
    if fx_obs.forecast.name == fx_obs.data_object.name:
        name += ' Forecast'
    return name


def _obs_color(interval_length):
    idx = np.searchsorted(OBS_PALETTE_TD_RANGE, interval_length)
    obs_color = OBS_PALETTE[idx]
    return obs_color


def _boolean_filter_indices_by_pair(value_cds, pair_index):
    return value_cds.data['pair_index'] == pair_index


def _extract_metadata_from_cds(metadata_cds, hash_, hash_key):
    first_row = np.argwhere(metadata_cds.data[hash_key] == hash_)[0][0]
    return {
        'pair_index': metadata_cds.data['pair_index'][first_row],
        'observation_name': metadata_cds.data['observation_name'][first_row],
        'forecast_name': metadata_cds.data['forecast_name'][first_row],
        'interval_label': metadata_cds.data['interval_label'][first_row],
        'observation_color': metadata_cds.data['observation_color'][first_row],
    }


def timeseries(timeseries_value_cds, timeseries_meta_cds,
               start, end, units, timezone='UTC'):
    """
    Timeseries plot of one or more forecasts and observations.

    Parameters
    ----------
    timeseries_value_cds: bokeh.models.ColumnDataSource
        ColumnDataSource of timeseries data. See :py:func:`solarforecastarbiter.reports.reoports.figures.construct_timeseries_cds` for format.
    timeseries_meta_cds: bokeh.models.ColumnDataSource
        ColumnDataSource of metadata for each Observation Forecast pair. See :py:func:`solarforecastarbiter.reports.reoports.figures.construct_timeseries_cds` for format.
    start : pandas.Timestamp
        Report start time
    end : pandas.Timestamp
        Report end time
    timezone : str
        Timezone consistent with the data in the obs_fx_cds.

    Returns
    -------
    fig : bokeh.plotting.figure
    """  # NOQA
    palette = cycle(PALETTE)

    fig = figure(
        sizing_mode='scale_width', plot_width=900, plot_height=300,
        x_range=(start, end), x_axis_type='datetime',
        tools='pan,xwheel_zoom,box_zoom,box_select,lasso_select,reset,save',
        name='timeseries')

    plotted_objects = 0
    for obs_hash in np.unique(timeseries_meta_cds.data['observation_hash']):
        metadata = _extract_metadata_from_cds(
            timeseries_meta_cds, obs_hash, 'observation_hash')
        pair_indices = _boolean_filter_indices_by_pair(
            timeseries_value_cds, metadata['pair_index'])
        view = CDSView(source=timeseries_value_cds, filters=[
            BooleanFilter(pair_indices)
        ])
        plot_method, plot_kwargs, hover_kwargs = line_or_step(
            metadata['interval_label'])
        legend_label = metadata['observation_name']
        color = metadata['observation_color']
        getattr(fig, plot_method)(
            x='timestamp', y='observation_values', source=timeseries_value_cds,
            view=view, color=color, legend_label=legend_label,
            **plot_kwargs)
        plotted_objects += 1

    for fx_hash in np.unique(timeseries_meta_cds.data['forecast_hash']):
        metadata = _extract_metadata_from_cds(
            timeseries_meta_cds, fx_hash, 'forecast_hash')
        pair_indices = _boolean_filter_indices_by_pair(
            timeseries_value_cds, metadata['pair_index'])
        view = CDSView(source=timeseries_value_cds,
                       filters=[BooleanFilter(pair_indices)])
        plot_method, plot_kwargs, hover_kwargs = line_or_step(
            metadata['interval_label'])
        legend_label = metadata['forecast_name']
        color = next(palette)
        getattr(fig, plot_method)(
            x='timestamp', y='forecast_values', source=timeseries_value_cds,
            view=view, color=color, legend_label=legend_label,
            **plot_kwargs)
        plotted_objects += 1

    fig.legend.location = "top_left"
    fig.legend.click_policy = "hide"
    if plotted_objects > 10:
        fig.legend.label_height = 10
        fig.legend.label_text_font_size = '8px'
        fig.legend.glyph_height = 10
        fig.legend.spacing = 1
        fig.legend.margin = 0
    fig.xaxis.axis_label = f'Time ({timezone})'
    fig.yaxis.axis_label = f'Data ({units})'
    return fig


def _get_scatter_limits(cds):
    extremes = [np.nan]
    for kind in ('forecast_values', 'observation_values'):
        arr = np.asarray(cds.data[kind]).astype(float)
        if len(arr) != 0:
            extremes.append(np.nanmin(arr))
            extremes.append(np.nanmax(arr))
    min_ = np.nanmin(extremes)
    if np.isnan(min_):
        min_ = -999
    max_ = np.nanmax(extremes)
    if np.isnan(max_):
        max_ = 999
    return min_, max_


def scatter(timeseries_value_cds, timeseries_meta_cds, units):
    """
    Scatter plot of one or more forecasts and observations.

    Parameters
    ----------
    timeseries_value_cds: bokeh.models.ColumnDataSource
        ColumnDataSource of timeseries data. See
        :py:func:`solarforecastarbiter.reports.reoports.figures.construct_timeseries_cds`
        for format.
    timeseries_meta_cds: bokeh.models.ColumnDataSource
        ColumnDataSource of metadata for each Observation Forecast pair. See
        :py:func:`solarforecastarbiter.reports.reoports.figures.construct_timeseries_cds`
        for format.

    Returns
    -------
    fig : bokeh.plotting.figure
    """  # NOQA
    xy_min, xy_max = _get_scatter_limits(timeseries_value_cds)

    # match_aspect=True does not work well, so these need to be close
    plot_height = 400
    # width will be updated later based on label length
    plot_width = plot_height + 50
    fig = figure(
        plot_width=plot_width, plot_height=plot_height, match_aspect=True,
        x_range=Range1d(xy_min, xy_max), y_range=Range1d(xy_min, xy_max),
        tools='pan,wheel_zoom,box_zoom,box_select,lasso_select,reset,save',
        name='scatter')

    kwargs = dict(size=6, line_color=None)

    palette = cycle(PALETTE)

    # accumulate labels and plot objects for manual legend
    scatters_labels = []
    for fxhash in np.unique(timeseries_meta_cds.data['forecast_hash']):
        metadata = _extract_metadata_from_cds(
            timeseries_meta_cds, fxhash, 'forecast_hash')
        pair_indices = _boolean_filter_indices_by_pair(
            timeseries_value_cds, metadata['pair_index'])
        view = CDSView(source=timeseries_value_cds,
                       filters=[BooleanFilter(pair_indices)])
        label = metadata['forecast_name']
        r = fig.scatter(
            x='observation_values', y='forecast_values',
            source=timeseries_value_cds, view=view,
            fill_color=next(palette), **kwargs)
        scatters_labels.append((label, [r]))

    # manual legend so it can be placed outside the plot area
    legend = Legend(items=scatters_labels, location='top_center',
                    click_policy='hide')
    fig.add_layout(legend, 'right')

    # compute new plot width accounting for legend label text width.
    # also considered using second figure for legend so it doesn't
    # distort the first when text length/size changes. unfortunately,
    # that doesn't work due to bokeh's inability to communicate legend
    # information across figures.
    # widest part of the legend
    max_legend_length = max((len(label) for label, _ in scatters_labels))
    px_per_length = 7.75  # found through trial and error
    fig.plot_width = int(fig.plot_width + max_legend_length * px_per_length)

    label = f'({units})'
    fig.xaxis.axis_label = 'Observed ' + label
    fig.yaxis.axis_label = 'Forecast ' + label
    return fig


def construct_metrics_cds(metrics, rename=None):
    """
    Possibly bad assumptions:
    * metrics contains keys: name, Total, etc.

    Parameters
    ----------
    metrics : list of datamodel.MetricResults
        Each metric dict is for a different forecast. Forecast name is
        specified by the name key.
    rename : function or None
        Function of one argument that is applied to each forecast name.

    Returns
    -------
    cds : bokeh.models.ColumnDataSource
        ColumnDataSource with indices 'name', 'abbrev', 'category', 'metric',
        and 'value'.
    """

    if rename:
        f = rename
    else:
        def f(x): return x  # NOQA

    data = []
    for metric_result in metrics:
        for mvalue in metric_result.values:
            new = {
                'name': metric_result.name,
                'abbrev': f(metric_result.name),
                'category': mvalue.category,
                'metric': mvalue.metric,
                'value': mvalue.value
            }
            if new['category'] == 'date':
                new['index'] = dt.datetime.strptime(
                    mvalue.index, '%Y-%m-%d')
            else:
                new['index'] = mvalue.index
            data.append(new)
    df = pd.DataFrame(data, columns=[
        'name', 'abbrev', 'category', 'metric', 'value', 'index'
    ])
    cds = ColumnDataSource(df, name='metrics_cds')
    cds.data.pop('level_0', None)
    return cds


def abbreviate(x, limit=3):
    # might need to add logic to ensure uniqueness
    # and/or enforce max length using textwrap.shorten
    components = x.split(' ')
    out_components = []
    for c in components:
        if len(c) <= limit:
            out = c
        elif c.upper() == c:
            # probably an acronym
            out = c
        else:
            out = f'{c[0:limit]}.'
        out_components.append(out)
    return ' '.join(out_components)


def bar(cds, metric):
    """
    Create a bar graph comparing a single metric across forecasts.

    Parameters
    ----------
    cds : bokeh.models.ColumnDataSource
        Metric cds created by :py:func:`solarforecastarbiter.reports.figures.construct_metrics_cds`
    metric: str
        The metric to plot. This value should be found in cds['metric'].

    Returns
    -------
    data_table : bokeh.widgets.DataTable
    """  # NOQA
    x_range = np.unique(cds.data['abbrev'])
    palette = cycle(PALETTE)
    palette = [next(palette) for _ in x_range]
    metric_name = datamodel.ALLOWED_METRICS[metric]
    view = CDSView(source=cds, filters=[
        GroupFilter(column_name='metric', group=metric),
        GroupFilter(column_name='category', group='total')
    ])
    # TODO: add units to title
    fig = figure(x_range=x_range, width=800, height=200, title=metric_name,
                 name=f'{metric}_total_bar', toolbar_location='above',
                 tools='pan,xwheel_zoom,box_zoom,reset,save')
    fig.vbar(x='abbrev', top='value', width=0.8,
             source=cds, view=view,
             line_color='white',
             fill_color=factor_cmap('abbrev', palette, factors=x_range))
    fig.xgrid.grid_line_color = None

    tooltips = [
        ('Forecast', '@name'),
        (metric_name, '@value'),
    ]
    hover = HoverTool(tooltips=tooltips, mode='vline')
    # more accurate would be if any single name is longer than each
    # name's allotted space. For example, never need to rotate labels
    # if forecasts are named A, B, C, D... but quickly need to rotate
    # if they have long names.
    if len(x_range) > 6:
        # pi/4 looks a lot better, but first tick label flows off chart
        # and I can't figure out how to add padding in bokeh
        fig.xaxis.major_label_orientation = np.pi / 2
        fig.width = 800
        # add more height to figure so that the names can go somewhere.
        fig.height = 400
    fig.add_tools(hover)
    return fig


def calc_y_start_end(y_min, y_max, pad_factor=1.03):
    """
    Determine y axis start, end.

    Parameters
    ----------
    y_min : float
    y_max : float
    pad_factor : float
        Number by which to multiply the start, end.

    Returns
    -------
    start, end : float, float
    """
    # bokeh does not play well with nans
    y_min = np.nan_to_num(y_min)
    y_max = np.nan_to_num(y_max)

    if y_max < 0:
        # all negative, so set range from y_min to 0
        start = y_min
        end = 0
    elif y_min > 0:
        # all positive, so set range from 0 to y_max
        start = 0
        end = y_max
    else:
        start = y_min
        end = y_max

    start, end = pad_factor * start, pad_factor * end
    return start, end


def bar_subdivisions(cds, category, metric):
    """
    Create bar graphs comparing a single metric across subdivisions of
    time for multiple forecasts. e.g.::

        Fx 1 MAE |
                 |_________________
        Fx 2 MAE |
                 |_________________
                   Year, Month of the year, etc.

    Parameters
    ----------
    cds : bokeh.models.ColumnDataSource
        Fields must be kind and the names of the forecasts
    category : str
        One of the available metrics grouping categories (e.g., total)

    Returns
    -------
    figs : dict of figures
    """
    palette = cycle(PALETTE)
    tools = 'pan,xwheel_zoom,box_zoom,reset,save'
    fig_kwargs = dict(tools=tools, toolbar_location='above')
    figs = {}

    width = 0.8

    human_category = datamodel.ALLOWED_CATEGORIES[category]
    metric_name = datamodel.ALLOWED_DETERMINISTIC_METRICS[metric]

    fig_kwargs['x_axis_label'] = human_category
    fig_kwargs['y_axis_label'] = metric_name

    filter_ = ((np.asarray(cds.data['category']) == category) &
               (np.asarray(cds.data['metric']) == metric))
    # Special handling for x-axis with dates
    if category == 'date':
        fig_kwargs['x_axis_type'] = 'datetime'
        width = width * pd.Timedelta(days=1)
        fig_kwargs['x_range'] = DataRange1d()
    elif category == 'month':
        fig_kwargs['x_range'] = FactorRange(
            factors=calendar.month_abbr[1:])
    elif category == 'weekday':
        fig_kwargs['x_range'] = FactorRange(
            factors=calendar.day_abbr[0:])
    elif category == 'hour':
        fig_kwargs['x_range'] = FactorRange(
            factors=[str(i) for i in range(25)])
    else:
        fig_kwargs['x_range'] = FactorRange(
            factors=np.unique(cds.data['index'][filter_]))

    y_data = np.asarray(cds.data['value'])[filter_]
    if len(y_data) == 0:
        start, end = None, None
    else:
        y_min = np.nanmin(y_data)
        y_max = np.nanmax(y_data)
        start, end = calc_y_start_end(y_min, y_max)
    fig_kwargs['y_range'] = DataRange1d(start=start, end=end)

    unique_names = np.unique(np.asarray(cds.data['name'])[filter_])

    for name in unique_names:
        view = CDSView(source=cds, filters=[
            GroupFilter(column_name='metric', group=metric),
            GroupFilter(column_name='category', group=category),
            GroupFilter(column_name='name', group=name)
        ])

        # Create figure
        title = name + ' ' + metric_name
        fig = figure(width=800, height=200, title=title,
                     name=f'{category}_{metric}_{name}',
                     **fig_kwargs)

        # Custom bar alignment
        if category == 'hour':
            # Center bars between hour ticks
            x = dodge('index', 0.5, range=fig.x_range)
        else:
            x = 'index'

        fig.vbar(x=x, top='value', width=width, source=cds,
                 view=view,
                 line_color='white', fill_color=next(palette))

        # axes parameters
        fig.xgrid.grid_line_color = None
        fig.xaxis.minor_tick_line_color = None

        # Hover tool and format specific changes
        if category == 'date':
            # Datetime x-axis
            formatter = DatetimeTickFormatter(days='%Y-%m-%d')
            fig.xaxis.formatter = formatter
            tooltips = [
                ('Forecast', '@name'),
                (human_category, '@index{%F}'),
                (metric_name, '@value'),
            ]
            hover_kwargs = dict(tooltips=tooltips,
                                formatters={'index': 'datetime'})
        elif category == 'month' or category == 'weekday':
            # Categorical x-axis
            formatter = CategoricalTickFormatter()
            fig.xaxis.formatter = formatter
            tooltips = [
                ('Forecast', '@name'),
                (human_category, '@index'),
                (metric_name, '@value'),
            ]
            hover_kwargs = dict(tooltips=tooltips)
        else:
            # Numerical x-axis
            tooltips = [
                ('Forecast', '@name'),
                (human_category, '@index'),
                (metric_name, '@value'),
            ]
            hover_kwargs = dict(tooltips=tooltips)
        hover = HoverTool(mode='vline', **hover_kwargs)
        fig.add_tools(hover)

        figs[name] = fig

    return figs


def nested_bar():
    raise NotImplementedError


def joint_distribution():
    raise NotImplementedError


def marginal_distribution():
    raise NotImplementedError


def taylor_diagram():
    raise NotImplementedError


def probabilistic_timeseries():
    raise NotImplementedError


def reliability_diagram():
    raise NotImplementedError


def rank_histogram():
    raise NotImplementedError


@contextmanager
def _make_webdriver():
    """Necessary until Bokeh 2.0 when using chrome/firefox drivers will be
    preferred and to avoid zombie phantomjs processes for now"""
    from selenium import webdriver
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        try:
            driver = webdriver.PhantomJS()
        except Exception:
            yield None
        else:
            yield driver
            driver.quit()


def output_svg(fig, driver=None):
    """
    Generates an SVG from the Bokeh figure. Errors in the
    process are logged and an SVG with error text is returned.

    Parameters
    ----------
    fig : bokeh.plotting.Figure
    driver : selenium.webdriver.remote.webdriver.WebDriver, default None
        Web driver to use to render SVG figures. With bokeh<2.0 this
        defaults to trying to use phantomjs.

    Returns
    -------
    svg : str
    """
    fig.output_backend = 'svg'
    try:
        svg = get_svgs(fig, driver=driver)[0]
    except Exception:
        logger.error('Could not generate SVG for figure %s',
                     getattr(fig, 'name', 'unnamed'))
        svg = (
            '<svg width="100%" height="100%">'
            '<text x="50" y="50" class="alert alert-error">'
            'Unable to generate SVG plot.'
            '</text>'
            '</svg>')
    return svg


def raw_report_plots(report, metrics):
    """Create a RawReportPlots object from the metrics of a report.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`
    metrics: tuple of :py:class:`solarforecastarbiter.datamodel.MetricResult`

    Returns
    -------
    :py:class:`solarforecastarbiter.datamodel.RawReportPlots`
    """
    cds = construct_metrics_cds(metrics, rename=abbreviate)
    # Create initial bar figures
    figure_dict = {}
    # Components for other metrics
    for category in report.report_parameters.categories:
        for metric in report.report_parameters.metrics:
            if category == 'total':
                fig = bar(cds, metric)
                figure_dict[f'total::{metric}::all'] = fig
            else:
                figs = bar_subdivisions(cds, category, metric)
                for name, fig in figs.items():
                    figure_dict[f'{category}::{metric}::{name}'] = fig
    script, divs = components(figure_dict)
    mplots = []

    with _make_webdriver() as driver:
        for k, v in divs.items():
            cat, met, name = k.split('::', 2)
            fig = figure_dict[k]
            svg = output_svg(fig, driver=driver)
            mplots.append(datamodel.BokehReportFigure(
                name=name, category=cat, metric=met, div=v, svg=svg,
                figure_type='bar'))

    out = datamodel.RawReportPlots(bokeh_version=bokeh_version, script=script,
                                   figures=tuple(mplots))
    return out


def timeseries_plots(report):
    """Return the bokeh components (script and div element) for timeseries
    and scatter plots of the processed forecasts and observations.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`

    Returns
    -------
    script: str
        A script element to insert into an html template
    div: str
        A div element to insert into an html template.
    """
    value_cds, meta_cds = construct_timeseries_cds(report)
    pfxobs = report.raw_report.processed_forecasts_observations
    units = pfxobs[0].original.forecast.units
    tfig = timeseries(value_cds, meta_cds, report.report_parameters.start,
                      report.report_parameters.end, units,
                      report.raw_report.timezone)
    sfig = scatter(value_cds, meta_cds, units)
    layout = gridplot((tfig, sfig), ncols=1)
    script, div = components(layout)
    return script, div



import os
import inspect
import pandas as pd
from solarforecastarbiter.io.fetch import eia


TEST_DATA_DIR = os.path.dirname(
    os.path.abspath(inspect.getfile(inspect.currentframe())))
GOOD_FILE = os.path.join(TEST_DATA_DIR, 'data', 'eia_caiso.json')
BAD_FILE = os.path.join(TEST_DATA_DIR, 'data', 'eia_caiso_bad.json')


def test_get_eia_data(requests_mock):
    with open(GOOD_FILE) as f:
        content = f.read()

    requests_mock.register_uri(
        "GET",
        "https://api.eia.gov/series/?api_key=fake_key&series_id=EBA.CISO-ALL.D.H&start=20200601T00Z&end=20200602T00Z",   # NOQA
        content=content.encode(),
    )

    series_id = "EBA.CISO-ALL.D.H"
    start = pd.Timestamp("2020-06-01 00:00Z")
    end = pd.Timestamp("2020-06-02 00:00Z")
    api_key = "fake_key"

    df = eia.get_eia_data(series_id, api_key, start, end)
    assert len(df) == 25
    assert df.index[0] == start
    assert df.index[-1] == end


def test_get_eia_data_bad(requests_mock):
    with open(BAD_FILE) as f:
        content = f.read()

    requests_mock.register_uri(
        "GET",
        "https://api.eia.gov/series/?api_key=fake_key&series_id=EBA.CISO-ALL.D.H&start=20200601T00Z&end=20200602T00Z",   # NOQA
        content=content.encode(),
    )

    series_id = "EBA.CISO-ALL.D.H"
    start = pd.Timestamp("2020-06-01 00:00Z")
    end = pd.Timestamp("2020-06-02 00:00Z")
    api_key = "fake_key"

    df = eia.get_eia_data(series_id, api_key, start, end)
    assert len(df) == 25
    assert df.index[0] == start
    assert df.index[-1] == end
    assert len(df.dropna()) == 22



import base64
import shutil


from plotly import graph_objects
import numpy as np
import pandas as pd
import pytest


import solarforecastarbiter.reports.figures.plotly_figures as figures
from solarforecastarbiter import datamodel


@pytest.fixture
def report_with_raw(report_dict, raw_report):
    report_dict['raw_report'] = raw_report(True)
    report = datamodel.Report.from_dict(report_dict)
    return report


@pytest.fixture
def report_with_raw_xy(report_dict, raw_report_xy):
    report_dict['raw_report'] = raw_report_xy(True)
    report = datamodel.Report.from_dict(report_dict)
    return report


@pytest.fixture
def raw_report_pfxobs_values(raw_report):
    # Useful for testing the valid types for a ProcessedForecastObservation's
    # values fields which can be either pd.Series, str or None
    def replace_values(value):
        raw = raw_report(False)
        raw = raw.replace(
            processed_forecasts_observations=tuple(
                pfxobs.replace(
                    forecast_values=value,
                    observation_values=value,
                )
                for pfxobs in raw.processed_forecasts_observations
            )
        )
        return raw
    return replace_values


@pytest.fixture
def set_report_pfxobs_values(report_dict, raw_report_pfxobs_values):
    def set_pfxobs_values(value):
        report_dict['raw_report'] = raw_report_pfxobs_values(value)
        report = datamodel.Report.from_dict(report_dict)
        return report
    return set_pfxobs_values


def test_construct_metrics_dataframe(report_with_raw):
    report = report_with_raw
    metrics = report.raw_report.metrics
    df = figures.construct_metrics_dataframe(metrics)
    names = df['name']
    abbrev = df['abbrev']
    categories = df['category']
    metrics = df['metric']
    values = df['value']
    report_params = report.report_parameters

    expected_length = (len(report_params.metrics) *
                       len(report_params.categories) *
                       len(report_params.object_pairs))
    assert all([len(v) == expected_length for k, v in df.items()])

    original_names = [fxobs.forecast.name
                      for fxobs in report_params.object_pairs]
    assert np.all(
        names == np.repeat(
            np.array(original_names),
            len(report_params.metrics) * len(report_params.categories))
    )
    assert np.all(names == abbrev)

    assert np.all(
        metrics == np.tile(
            np.repeat(np.array(report_params.metrics, dtype=object),
                      len(report_params.categories)),
            len(report_params.object_pairs))
    )

    assert np.all(
        categories == np.tile(
            np.array(report_params.categories),
            len(report_params.metrics) * len(report_params.object_pairs))
    )

    # this could maybe use value variance, but asserting the dataframe process
    # did not mangle values for now
    assert (values == 2).all()


def test_construct_metrics_dataframe_with_rename(report_with_raw):
    metrics = report_with_raw.raw_report.metrics
    df = figures.construct_metrics_dataframe(metrics,
                                             rename=figures.abbreviate)
    report_params = report_with_raw.report_parameters
    original_names = [fxobs.forecast.name
                      for fxobs in report_params.object_pairs]
    abbreviated = list(map(figures.abbreviate, original_names))
    assert np.all(
        df['abbrev'] == np.repeat(
            np.array(abbreviated, dtype=object),
            len(report_params.metrics) * len(report_params.categories))
    )


def test_construct_metric_dataframe_no_values():
    # Iterative metrics datafame creation just builds an empty dataframe
    # with correct columns if no MetricResults are found in the metrics tuple
    df = figures.construct_metrics_dataframe(())
    assert df['index'].size == 0
    assert 'abbrev' in df


@pytest.fixture
def fxobs_name_mock(mocker):
    def fn(obs_name, fx_name, agg=False, fxspec=None):
        if agg:
            obspec = datamodel.Aggregate
        else:
            obspec = datamodel.Observation
        fxobs = mocker.Mock()
        obs = mocker.Mock(spec=obspec)
        obs.name = obs_name
        fx = mocker.Mock(spec=fxspec)
        fx.name = fx_name
        fxobs.forecast = fx
        fxobs.data_object = obs
        return fxobs
    return fn


@pytest.mark.parametrize('obs_name,fx_name,fx_name_expected', [
    ('obs', 'fx', 'fx'),
    ('name', 'name', 'name Forecast')
])
def test__fx_name(fxobs_name_mock, obs_name, fx_name, fx_name_expected):
    fxobs = fxobs_name_mock(obs_name, fx_name)
    fx_name = figures._fx_name(fxobs.forecast, fxobs.data_object)
    assert fx_name == fx_name_expected


@pytest.mark.parametrize('axis,fx_name_expected', [
    ('x', 'fx name Prob(x <= 10.0 MW)'),
    ('y', 'fx name Prob(f <= x) = 10.0%')
])
def test__fx_name_prob(axis, fx_name_expected, fxobs_name_mock):
    fxobs = fxobs_name_mock(
        'obs name', 'fx name',
        fxspec=datamodel.ProbabilisticForecastConstantValue)
    fxobs.forecast.constant_value = 10.
    fxobs.forecast.axis = axis
    fxobs.forecast.units = 'MW'
    fx_name = figures._fx_name(fxobs.forecast, fxobs.data_object)
    assert fx_name == fx_name_expected


@pytest.mark.parametrize('y_min,y_max,pad,expected', [
    (1.0, 10.0, 0.5, (0.0, 5.0)),
    (-5.0, 10.0, 1.5, (-7.5, 15.0)),
    (-5.0, -1.0, 1.0, (-5.0, 0.0)),
    (-100.0, 100.0, 1.03, (-103.0, 103.0)),
])
def test_start_end(y_min, y_max, pad, expected):
    result = figures.calc_y_start_end(y_min, y_max, pad)
    assert result == expected


@pytest.mark.parametrize('arg,expected', [
    ('a', 'a'),
    ('abcd', 'abc.'),
    ('ABCDEFGHIJKLMNOP', 'ABCDEFGHIJKLMNOP'),
    ('University of Arizona OASIS Day Ahead GFS ghi',
     'Uni. of Ari. OASIS Day Ahe. GFS ghi')
])
def test_abbreviate(arg, expected):
    out = figures.abbreviate(arg)
    assert out == expected


def test_raw_report_plots(report_with_raw):
    metrics = report_with_raw.raw_report.metrics
    plots = figures.raw_report_plots(report_with_raw, metrics)
    assert plots is not None


def test_output_svg_with_plotly_figure(mocker):
    logger = mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.logger')
    if shutil.which('orca') is None:  # pragma: no cover
        pytest.skip('orca must be on PATH to make SVGs')
    if shutil.which('xvfb-run') is None:  # pragma: no cover
        pytest.skip('xvfb-run must be on PATH to make SVGs')
    values = list(range(5))
    fig = graph_objects.Figure(data=graph_objects.Scatter(x=values, y=values))
    svg = figures.output_svg(fig)
    assert svg.startswith('<svg')
    assert svg.endswith('</svg>')
    assert not logger.error.called


def test_output_pdf_with_plotly_figure(mocker):
    logger = mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.logger')
    if shutil.which('orca') is None:  # pragma: no cover
        pytest.skip('orca must be on PATH to make PDFs')
    if shutil.which('xvfb-run') is None:  # pragma: no cover
        pytest.skip('xvfb-run must be on PATH to make PDFs')
    values = list(range(5))
    fig = graph_objects.Figure(data=graph_objects.Scatter(x=values, y=values))
    pdf = figures.output_pdf(fig)
    pdf_bytes = base64.a85decode(pdf)
    assert pdf_bytes.startswith(b'%PDF-')
    assert pdf_bytes.rstrip(b'\n').endswith(b'%%EOF')
    assert not logger.error.called


def test_output_svg_with_plotly_figure_no_orca(mocker, remove_orca):
    logger = mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.logger')
    values = list(range(5))
    fig = graph_objects.Figure(data=graph_objects.Scatter(x=values, y=values))
    svg = figures.output_svg(fig)
    assert svg.startswith('<svg')
    assert 'Unable' in svg
    assert svg.endswith('</svg>')
    assert logger.error.called


def test_output_pdf_with_plotly_figure_no_orca(mocker, remove_orca):
    logger = mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.logger')
    values = list(range(5))
    fig = graph_objects.Figure(data=graph_objects.Scatter(x=values, y=values))
    pdf = figures.output_pdf(fig)
    pdf_bytes = base64.a85decode(pdf)
    assert pdf_bytes.startswith(b'%PDF-')
    assert pdf_bytes.rstrip(b'\n').endswith(b'%%EOF')
    assert logger.error.called


@pytest.fixture()
def metric_dataframe():
    return pd.DataFrame({
        'name': ['First', 'Next'],
        'abbrev': ['1st', 'N'],
        'category': ['hour', 'total'],
        'metric': ['mae', 'mae'],
        'value': [0.1, 10.3],
        'index': [14, 0]
    })


def test_bar(metric_dataframe):
    out = figures.bar(metric_dataframe, 'mae')
    assert isinstance(out, graph_objects.Figure)


def test_bar_no_metric(metric_dataframe):
    out = figures.bar(metric_dataframe, 'rmse')
    assert isinstance(out, graph_objects.Figure)


def test_bar_empty_df(metric_dataframe):
    df = pd.DataFrame(columns=metric_dataframe.columns)
    out = figures.bar(df, 's')
    assert isinstance(out, graph_objects.Figure)


def test_bar_subdivisions(metric_dataframe):
    out = figures.bar_subdivisions(metric_dataframe, 'hour', 'mae')
    assert isinstance(out, dict)
    assert len(out) == 1
    assert all([isinstance(v, graph_objects.Figure) for v in out.values()])


def test_bar_subdivisions_no_cat(metric_dataframe):
    out = figures.bar_subdivisions(metric_dataframe, 'date', 'mae')
    assert isinstance(out, dict)
    assert len(out) == 0


def test_bar_subdivisions_no_metric(metric_dataframe):
    out = figures.bar_subdivisions(metric_dataframe, 'hour', 'rmse')
    assert isinstance(out, dict)
    assert len(out) == 0


def test_bar_subdivisions_empty_df(metric_dataframe):
    df = pd.DataFrame(columns=metric_dataframe.columns)
    out = figures.bar_subdivisions(df, 'hour', 's')
    assert isinstance(out, dict)
    assert len(out) == 0


@pytest.mark.parametrize('input_str,expected', [
    ('not long', 'not long'),
    ('pretty long name and stuff', 'pretty long name and<br>stuff'),
    ('long word in middlemiddlemiddlemiddle of it',
     'long word in<br>middlemiddlemiddlemiddle<br>of it'),
])
def test_legend_text(input_str, expected):
    assert figures._legend_text(input_str) == expected


ts_df_data = [
    {'timestamp': '2020-01-01T00:00Z',
     'value': 5},
    {'timestamp': '2020-01-01T00:05Z',
     'value': 5},
    {'timestamp': '2020-01-01T00:10Z',
     'value': 5},
    {'timestamp': '2020-01-01T00:25Z',
     'value': 5},
    {'timestamp': '2020-01-01T00:30Z',
     'value': 5},
]


def test_fill_timeseries():
    data = pd.DataFrame(ts_df_data)
    data = data.set_index('timestamp')
    filled = figures._fill_timeseries(data, np.timedelta64(5, 'm'))
    assert filled.index.size == 7
    assert pd.isnull(filled.iloc[-4].value)
    assert pd.isnull(filled.iloc[-3].value)


def test_fill_timeseries_empty():
    data = pd.DataFrame()
    filled = figures._fill_timeseries(data, np.timedelta64(5, 'm'))
    assert filled.empty


meta_df_data = [
    {
        'pair_index': 0,
        'observation_name': 'obs one',
        'forecast_name': 'fx one',
        'observation_hash': str(hash('obs one')),
        'forecast_hash': str(hash('fx one')),
        'interval_label': 'beginning',
        'interval_length': np.timedelta64(1, 'm'),
        'observation_color': '#abc',
        'forecast_type': None,
        'axis': None,
        'constant_value': None
    }, {
        'pair_index': 1,
        'observation_name': 'obs two',
        'forecast_name': 'fx two',
        'observation_hash': str(hash('obs two')),
        'forecast_hash': str(hash('fx two')),
        'interval_label': 'beginning',
        'interval_length': np.timedelta64(5, 'm'),
        'observation_color': '#ccc',
        'forecast_type': None,
        'axis': None,
        'constant_value': None
    },
]


@pytest.fixture(params=[0, 1])
def meta_entries(request):
    return meta_df_data[request.param]


@pytest.mark.parametrize('hash_key', [
    'observation_hash', 'forecast_hash'
])
def test_extract_metadata(hash_key, meta_entries):
    meta_df = pd.DataFrame(meta_df_data)
    extracted = figures._extract_metadata_from_df(
        meta_df, meta_entries[hash_key], hash_key)
    for k, v in extracted.items():
        assert meta_entries[k] == v


def test_construct_timeseries_dataframe_timeseries(report_with_raw):
    ts_df, _ = figures.construct_timeseries_dataframe(report_with_raw)
    assert str(ts_df.index.tz) == report_with_raw.raw_report.timezone
    assert (ts_df.count() == 288).all()
    assert list(np.unique(ts_df['pair_index'])) == [0, 1, 2]
    assert (ts_df['observation_values'] == 100).all()
    assert (ts_df['forecast_values'] == 100).all()


def test_timeseries_plots(report_with_raw):
    ts_spec, scatter_spec, ts_prob_spec, inc_dist = figures.timeseries_plots(
        report_with_raw)
    assert isinstance(ts_spec, str)
    assert isinstance(scatter_spec, str)
    assert ts_prob_spec is None
    assert not inc_dist


def test_timeseries_plots_xy(report_with_raw_xy):
    ts_spec, scatter_spec, ts_prob_spec, inc_dist = figures.timeseries_plots(
        report_with_raw_xy)
    assert isinstance(ts_spec, str)
    assert isinstance(scatter_spec, str)
    assert isinstance(ts_prob_spec, str)
    assert inc_dist


@pytest.fixture
def report_with_raw_xy_asymmetric_cv(report_dict, raw_report_xy):
    raw = raw_report_xy(True)
    pfxobs = raw.processed_forecasts_observations
    fx_cvs = pfxobs[0].original.forecast.constant_values
    cv_1 = pfxobs[0].replace(
        original=pfxobs[0].original.replace(
            forecast=pfxobs[0].original.forecast.replace(
                constant_values=(
                    fx_cvs[0],
                    fx_cvs[1],
                    fx_cvs[2].replace(constant_value=85.0))
            ),
        ),
        forecast_values=pfxobs[0].forecast_values.rename(
            columns={'75.0': '85.0'})
    )
    raw = raw.replace(
        processed_forecasts_observations=(cv_1, pfxobs[1], pfxobs[2])
    )
    report_dict['raw_report'] = raw
    return datamodel.Report.from_dict(report_dict)


def test_probabilistic_plotting_asymmetric_cv(
        report_with_raw_xy_asymmetric_cv):
    ts_spec, scatter_spec, ts_prob_spec, inc_dist = figures.timeseries_plots(
        report_with_raw_xy_asymmetric_cv)
    assert isinstance(ts_spec, str)
    assert isinstance(scatter_spec, str)
    assert isinstance(ts_prob_spec, str)
    assert inc_dist



"""
Functions for forecasting.
"""
import datetime

import pandas as pd
import numpy as np

from solarforecastarbiter import pvmodel

# some functions originally implemented in pvlib-python.
# find pvlib-python license in licenses directory.


def cloud_cover_to_ghi_linear(cloud_cover, ghi_clear, offset=35):
    """
    Convert cloud cover to GHI using a linear relationship.

    0% cloud cover returns ghi_clear.

    100% cloud cover returns offset*ghi_clear.

    Parameters
    ----------
    cloud_cover: numeric
        Cloud cover in %.
    ghi_clear: numeric
        GHI under clear sky conditions.
    offset: numeric, default 35
        Determines the minimum GHI.

    Returns
    -------
    ghi: numeric
        Cloudy sky GHI.

    References
    ----------
    Larson et. al. "Day-ahead forecasting of solar power output from
    photovoltaic plants in the American Southwest" Renewable Energy
    91, 11-20 (2016).
    """

    offset = offset / 100.
    cloud_cover = cloud_cover / 100.
    ghi = (offset + (1 - offset) * (1 - cloud_cover)) * ghi_clear
    return ghi


def cloud_cover_to_irradiance_ghi_clear(cloud_cover, ghi_clear, zenith):
    """
    Estimates irradiance from cloud cover in the following steps:

    1. Estimate cloudy sky GHI using a function of cloud_cover and
       ghi_clear: :py:func:`cloud_cover_to_ghi_linear`
    2. Estimate cloudy sky DNI and DHI using the Erbs model.

    Parameters
    ----------
    site : datamodel.Site
    cloud_cover : Series
        Cloud cover in %.
    zenith : Series
        Solar zenith

    Returns
    -------
    ghi : pd.Series, dni : pd.Series, dhi : pd.Series
    """
    ghi = cloud_cover_to_ghi_linear(cloud_cover, ghi_clear)
    dni, dhi = pvmodel.complete_irradiance_components(ghi, zenith)
    return ghi, dni, dhi


def cloud_cover_to_irradiance(latitude, longitude, elevation, cloud_cover,
                              apparent_zenith, zenith):
    """
    Estimates irradiance from cloud cover in the following steps:

    1. Determine clear sky GHI using Ineichen model and climatological
       turbidity.
    2. Estimate cloudy sky GHI using a function of cloud_cover and
       ghi_clear: :py:func:`cloud_cover_to_ghi_linear`
    3. Estimate cloudy sky DNI and DHI using the Erbs model.

    Don't use this function if you already have clear sky GHI. Instead,
    use :py:func:`cloud_cover_to_irradiance_ghi_clear`

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    cloud_cover : Series
        Cloud cover in %.
    apparent_zenith : Series
        Solar apparent zenith
    zenith : Series
        Solar zenith

    Returns
    -------
    ghi : pd.Series
    dni : pd.Series
    dhi : pd.Series

    See also
    --------
    cloud_cover_to_irradiance_ghi_clear
    cloud_cover_to_ghi_linear
    """
    cs = pvmodel.calculate_clearsky(latitude, longitude, elevation,
                                    apparent_zenith)
    ghi, dni, dhi = cloud_cover_to_irradiance_ghi_clear(
        cloud_cover, cs['ghi'], zenith)
    return ghi, dni, dhi


def resample(arg, freq='1h', label=None):
    """Resamples an argument, allowing for None. Use with map.
    Useful for applying resample to unknown model output (e.g. AC power
    is None if no plant metadata is provided).

    Parameters
    ----------
    arg : pd.Series or None
    freq : str
    label : str
        Sets pandas resample's label and closed kwargs.

    Returns
    -------
    pd.Series or None
    """
    if arg is None:
        return None
    else:
        return arg.resample(freq, label=label, closed=label).mean()


def reindex_fill_slice(arg, freq='5min', label=None, start=None, end=None,
                       start_slice=None, end_slice=None,
                       fill_method='interpolate'):
    """Reindex data to shorter intervals (create NaNs) from `start` to
    `end`, fill NaNs in gaps using `fill_method`, fill NaNs before first
    valid time using bfill, fill NaNs after last valid time using ffill,
    then slice output from `start_slice` to `end_slice`.
    """
    if arg is None or arg.empty:
        return arg
    if start is None:
        start_reindex = arg.index[0]
    else:
        start_reindex = min(pd.Timestamp(start), arg.index[0])
    if end is None:
        end_reindex = arg.index[-1]
    else:
        end_reindex = max(pd.Timestamp(end), arg.index[-1])
    index = pd.date_range(start=start_reindex, end=end_reindex, freq=freq,
                          closed=label)
    reindexed = arg.reindex(index)
    filled = getattr(reindexed, fill_method)()
    sliced = filled.loc[start_slice:end_slice].bfill().ffill()
    return sliced


def unmix_intervals(mixed, lower=0, upper=100):
    """Convert mixed interval averages into pure interval averages.

    For example, the GFS 3 hour output contains the following data:

    * forecast hour 3: average cloud cover from 0 - 3 hours
    * forecast hour 6: average cloud cover from 0 - 6 hours
    * forecast hour 9: average cloud cover from 6 - 9 hours
    * forecast hour 12: average cloud cover from 6 - 12 hours

    and so on. This function returns:

    * forecast hour 3: average cloud cover from 0 - 3 hours
    * forecast hour 6: average cloud cover from 3 - 6 hours
    * forecast hour 9: average cloud cover from 6 - 9 hours
    * forecast hour 12: average cloud cover from 9 - 12 hours

    Parameters
    ----------
    data : pd.Series
        The first time must be the first output of a cycle.
    lower : None or float
        Lower bound. Useful for handling numerical precision issues in
        input data.
    upper : None or float
        Upper bound of output. Useful for handling numerical precision
        issues in input data.

    Returns
    -------
    pd.Series
        Data is the unmixed interval average with ending label.
    """
    intervals = (mixed.index[1:] - mixed.index[:-1]).unique()
    if len(intervals) > 1:
        raise ValueError('multiple interval lengths detected. slice forecasts '
                         'into sections with unique interval lengths first.')
    interval = intervals[0]
    start = mixed.index[0]
    mixed_vals = np.array(mixed)
    if interval == pd.Timedelta('1h'):
        _check_start_time(start, interval)
        # mixed_1...mixed_6 are the values in the raw forecast file at
        # each hour of the mixed interval period. Let f1...f6 be unmixed,
        # true hourly average values. The relationship is:
        # mixed_1 = f1
        # mixed_2 = (f1 + f2) / 2
        # mixed_3 = (f1 + f2 + f3) / 3
        # mixed_4 = (f1 + f2 + f3 + f4) / 4
        # mixed_5 = (f1 + f2 + f3 + f4 + f5) / 5
        # mixed_6 = (f1 + f2 + f3 + f4 + f5 + f6) / 6
        # some algebra will show that the f1...f6 can be obtained as
        # coded below.
        # the cycle repeats itself after 6 hours so
        # mixed_7 = f7
        # mixed_8 = (f8 + f9) / 2 ...
        # To efficiently compute the values for all forecast times,
        # we use slices for every 6th element, calculate the forecasts
        # at every 6th point, then interleave them by constructing a 2D
        # array and reshaping it to a 1D array.
        mixed_1 = mixed_vals[0::6]
        mixed_2 = mixed_vals[1::6]
        mixed_3 = mixed_vals[2::6]
        mixed_4 = mixed_vals[3::6]
        mixed_5 = mixed_vals[4::6]
        mixed_6 = mixed_vals[5::6]
        f1 = mixed_1
        f2 = 2 * mixed_2 - mixed_1
        f3 = 3 * mixed_3 - 2 * mixed_2
        f4 = 4 * mixed_4 - 3 * mixed_3
        f5 = 5 * mixed_5 - 4 * mixed_4
        f6 = 6 * mixed_6 - 5 * mixed_5
        f = np.array([f1, f2, f3, f4, f5, f6])
    elif interval == pd.Timedelta('3h'):
        _check_start_time(start, interval)
        # similar to above, but
        # mixed_3 = f_0_3
        # mixed_6 = (f_0_3 + f_3_6) / 2
        f3 = mixed_vals[0::2]
        f6 = 2 * mixed_vals[1::2] - f3
        f = np.array([f3, f6])
    else:
        raise ValueError('mixed period must be 6 hours and data interval must '
                         'be 3 hours or 1 hour')
    unmixed = pd.Series(f.reshape(-1, order='F'), index=mixed.index)
    unmixed = unmixed.clip(lower=0, upper=100)
    return unmixed


def _check_start_time(start, interval):
    """
    start : pd.Timestamp
        Must be timezone aware.
    interval : pd.Timedelta
        Time between data points.
    """
    # for this to work, the first value must belong to the
    # first time of a mixed interval period. Assuming GFS data...
    start_time = start.tz_convert('UTC').time()
    allowed_times_interval = {
        pd.Timedelta('1h'): (1, 7, 13, 19),
        pd.Timedelta('3h'): (3, 9, 15, 21)
    }
    allowed_times = allowed_times_interval[interval]
    if any(start_time == datetime.time(t) for t in allowed_times):
        return
    else:
        raise ValueError(f'for {interval} mixed intervals, start time must '
                         f'be one of {allowed_times}Z hours')


def sort_gefs_frame(frame):
    """Sort a DataFrame from a GEFS forecast. Column 0 is the smallest
    value at each time. Column 20 is the largest value at each time.
    """
    if frame is None:
        return frame
    else:
        return pd.DataFrame(np.sort(frame), index=frame.index)



"""Functions to read NREL PVDAQ data.
"""

# Code originally written by Bennet Meyers (@bmeyers), Stanford, SLAC in
# https://github.com/pvlib/pvlib-python/pull/664
# Adapated by Will Holmgren (@wholmgren), University of Arizona

import json
from io import StringIO

import requests
import pandas as pd


# consider adding an auth=(username, password) kwarg (default None) to
# support private data queries

def get_pvdaq_metadata(system_id, api_key):
    """Query PV system metadata from NREL's PVDAQ data service.

    Parameters
    ----------
    system_id: int
        The system ID corresponding to the site that data should be
        queried from.

    api_key: string
        Your NREL API key (https://developer.nrel.gov/docs/api-key/)

    Returns
    -------
    list of dict
    """

    params = {'system_id': system_id, 'api_key': api_key}
    sites_url = 'https://developer.nrel.gov/api/pvdaq/v3/sites.json'
    r = requests.get(sites_url, params=params)
    r.raise_for_status()
    outputs = json.loads(r.content)['outputs']
    return outputs


def get_pvdaq_data(system_id, year, api_key='DEMO_KEY'):
    """Query PV system data from NREL's PVDAQ data service:

    https://maps.nrel.gov/pvdaq/

    This function uses the annual raw data file API, which is the most
    efficient way of accessing multi-year, sub-hourly time series data.

    Parameters
    ----------
    system_id: int
        The system ID corresponding to the site that data should be
        queried from.

    year: int or list of ints
        Either the year to request or the list of years to request.
        Multiple years will be concatenated into a single DataFrame.

    api_key: string
        Your NREL API key (https://developer.nrel.gov/docs/api-key/)

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the time series data from the
        PVDAQ service over the years requested. Times are typically
        in local time.

    Notes
    -----
    The PVDAQ metadata contains a key "available_years" that is a useful
    value for the *year* argument.
    """

    try:
        year = int(year)
    except TypeError:
        year = [int(yr) for yr in year]
    else:
        year = [year]

    # Each year must queries separately, so iterate over the years and
    # generate a list of dataframes.
    # Consider putting this loop in its own private function with
    # try / except / try again pattern for network issues and NREL API
    # throttling
    df_list = []
    for yr in year:
        params = {
            'api_key': api_key,
            'system_id': system_id,
            'year': yr
        }
        base_url = 'https://developer.nrel.gov/api/pvdaq/v3/data_file'
        response = requests.get(base_url, params=params)
        response.raise_for_status()
        df = pd.read_csv(StringIO(response.text))
        df_list.append(df)

    # concatenate the list of yearly DataFrames
    df = pd.concat(df_list, axis=0, sort=True)
    df['Date-Time'] = pd.to_datetime(df['Date-Time'])
    df.set_index('Date-Time', inplace=True)
    return df



import pytest
import numpy as np
from solarforecastarbiter.metrics import summary


@pytest.mark.parametrize("ts", [
    [1, 2, 3],
    np.random.rand(10),
    np.random.rand(1000),
])
def test_scalar(ts):
    for metric in summary._MAP:
        f = summary._MAP[metric][0]
        assert np.isscalar(f(ts))



import logging


import pandas as pd
from pvlib import iotools


from solarforecastarbiter.io.reference_observations import (
    common, midc_config, default_forecasts)


logger = logging.getLogger('reference_data')


def initialize_site_observations(api, site):
    """Creates an observation at the site for each variable in the MIDC
    site's file.

    Parameters
    ----------
    api : io.api.APISession
        API Session object, authenticated for the Reference user.
    site : datamodel.Site
        The site object for which to create the Observations.
    """
    try:
        extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.warning('Cannot create reference observations at MIDC site '
                       f'{site.name}, missing required parameters.')
        return
    site_api_id = extra_params['network_api_id']
    for sfa_var, midc_var in midc_config.midc_var_map[site_api_id].items():
        obs_extra_params = extra_params.copy()
        obs_extra_params['network_data_label'] = midc_var
        common.create_observation(api, site, sfa_var,
                                  extra_params=obs_extra_params)


def initialize_site_forecasts(api, site):
    """
    Create a forecasts for each variable measured at the site

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Forecasts.
    """
    try:
        extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.warning('Cannot create reference observations at MIDC site '
                       f'{site.name}, missing required parameters.')
        return
    site_api_id = extra_params['network_api_id']
    common.create_forecasts(
        api, site, midc_config.midc_var_map[site_api_id].keys(),
        default_forecasts.TEMPLATE_FORECASTS)


def fetch(api, site, start, end):
    """Retrieve observation data for a MIDC site between start and end.

    Parameters
    ----------
    api : io.APISession
        Unused but conforms to common.update_site_observations call
    site : datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.
    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        return pd.DataFrame()
    try:
        obs_df = iotools.read_midc_raw_data_from_nrel(
            site_extra_params['network_api_id'], start, end)
    except IndexError:
        # The MIDC api returns a text file on 404 that is parsed as a
        # 1-column csv and causes an index error.
        logger.warning(f'Could not retrieve data for site {site.name}'
                       f' between {start} and {end}.')
        return pd.DataFrame()
    except ValueError as e:
        logger.error(f'Error retrieving data for site {site.name}'
                     f' between {start} and {end}: %s', e)
        return pd.DataFrame()
    return obs_df


def update_observation_data(api, sites, observations, start, end):
    """Post new observation data to all MIDC observations from
    start to end.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites: list
        List of all reference sites as Objects
    observations: list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    """
    midc_sites = common.filter_by_networks(sites, ['NREL MIDC'])
    for site in midc_sites:
        common.update_site_observations(
            api, fetch, site, observations, start, end)



"""
Make a report.

Steps:

  1. Consume metadata defined in :py:mod:`~solarforecastarbiter.datamodel`
  2. Run validation of metadata. Metadata creation might also include
     some validation so this may not be necessary.
  3. Get data using io.api.
  4. Align observation data to forecast data using metrics subpackage.
  5. Compute metrics specified in metadata using metrics subpackage
  6. Assemble metrics and aligned data into a raw report object which
     can then later be converted to a HTML or PDF report
  7. Prepare to post raw report, metrics, and aligned data to the API.
     The raw report sent to the API be JSON with keys like:
        * metadata
            * checksum of data
            * date created
            * versions
        * data: base64 encoded raw report

      The metrics will also be JSON with keys like:
        * metrics
            * total
            * category A
                * metric 1
                * metric 2
            * category B
                * metric 1
                * metric 2
            * category C
                * metric 1
                * metric 2
       The JSON could be large and difficult for a person to read, but
       should be relatively easy to parse.
       The aligned data will be encoded separately in a binary format.
  7. Submit the raw report to API.
  8. When the report is later requested from the API, get the raw report,
     metrics, and aligned data and convert to HTML or PDF


Considerations:

* API uses queue system to initiate report generation
* Functions should not require an API session unless they really need it.
* The bokeh plots in the html version will be rendered at
  client load time. The metrics data will be immediately available, but
  the API will need to call for the aligned data separately
  to be able to create time series, scatter, etc. plots.
"""
from functools import wraps
import pkg_resources
import platform


import pandas as pd


from solarforecastarbiter.io.api import APISession
from solarforecastarbiter import datamodel
from solarforecastarbiter.metrics import preprocessing, calculator
from solarforecastarbiter.reports.figures import plotly_figures
from solarforecastarbiter.utils import hijack_loggers
from solarforecastarbiter.validation.tasks import apply_validation


def get_data_for_report(session, report):
    """
    Get data for report.

    1 API call is made for each unique forecast and observation object.

    Parameters
    ----------
    session : :py:class:`solarforecastarbiter.api.APISession`
        API session for getting and posting data
    report : :py:class:`solarforecastarbiter.datamodel.Report`
        Metadata describing report

    Returns
    -------
    data : dict
        Keys are Forecast and Observation objects, values are
        the corresponding data. Keys also include any reference
        forecasts that exist in the report.
    """
    data = {}
    start = report.report_parameters.start
    end = report.report_parameters.end
    for fxobs in report.report_parameters.object_pairs:
        # forecasts and especially observations may be repeated.
        # only get the raw data once.
        if fxobs.forecast not in data:
            # use get_values instead of get_forecast_values so that api module
            # can handle determ., prob constant value, or prob group values
            data[fxobs.forecast] = session.get_values(
                fxobs.forecast, start, end)
        if fxobs.data_object not in data:
            obs_data = session.get_values(
                fxobs.data_object, start, end)
            data[fxobs.data_object] = apply_validation(
                fxobs.data_object, obs_data)
        if fxobs.reference_forecast is not None:
            if fxobs.reference_forecast not in data:
                data[fxobs.reference_forecast] = session.get_values(
                    fxobs.reference_forecast, start, end)

    return data


def get_versions():
    packages = [
        'solarforecastarbiter',
        'pvlib',
        'pandas',
        'numpy',
        'scipy',
        'statsmodels',
        'plotly',
        'bokeh',
        'netcdf4',
        'xarray',
        'tables',
        'numexpr',
        'bottleneck',
        'jinja2',
    ]
    versions = []
    for p in packages:
        try:
            v = pkg_resources.get_distribution(p).version
        except pkg_resources.DistributionNotFound:
            v = 'None'
        versions.append((p, str(v)))
    versions.append(('python', str(platform.python_version())))
    versions.append(('platform', platform.platform()))
    return tuple(versions)


def infer_timezone(report_parameters):
    # maybe not ideal when comparing across sites. might need explicit
    # tz options ('infer' or spec IANA tz) in report interface.
    fxobs_0 = report_parameters.object_pairs[0]
    if isinstance(fxobs_0, datamodel.ForecastObservation):
        timezone = fxobs_0.observation.site.timezone
    else:
        timezone = fxobs_0.aggregate.timezone
    return timezone


def create_raw_report_from_data(report, data):
    """
    Create a raw report using data and report metadata.

    Parameters
    ----------
    report : :py:class:`solarforecastarbiter.datamodel.Report`
        Metadata describing report
    data : dict
        Keys are all Forecast and Observation (or Aggregate)
        objects in the report, values are the corresponding data.

    Returns
    -------
    raw_report : :py:class:`solarforecastarbiterdatamodel.RawReport`

    Todo
    ----
    * add reference forecast
    """
    generated_at = pd.Timestamp.now(tz='UTC')
    report_params = report.report_parameters
    timezone = infer_timezone(report_params)
    versions = get_versions()
    with hijack_loggers([
        'solarforecastarbiter.metrics',
        'solarforecastarbiter.reports.figures.plotly_figures'],
                        ) as handler:
        # Validate, fill forecast, and resample
        processed_fxobs = preprocessing.process_forecast_observations(
            report_params.object_pairs,
            report_params.filters,
            report_params.forecast_fill_method,
            report_params.start, report_params.end,
            data, timezone,
            costs=report_params.costs)

        # Calculate metrics
        metrics_list = calculator.calculate_metrics(
            processed_fxobs,
            list(report_params.categories),
            list(report_params.metrics))

        report_plots = plotly_figures.raw_report_plots(report, metrics_list)
        messages = handler.export_records()
    raw_report = datamodel.RawReport(
        generated_at=generated_at, timezone=timezone, versions=versions,
        plots=report_plots, metrics=tuple(metrics_list),
        processed_forecasts_observations=tuple(processed_fxobs),
        messages=messages)
    return raw_report


def capture_report_failure(report_id, session):
    """
    Decorator factory to handle errors in report generation by
    posting a message in an empty RawReport along with a failed
    status to the API.

    Parameters
    ----------
    report_id: str
        ID of the report to update with the message and failed status
    session: :py:class:`solarforecastarbiter.io.api.APISession`
        Session object to connect to the API.

    Returns
    -------
    decorator
        Decorator to handle any errors in the decorated function.
        The decorator has an optional `err_msg` keyword argument
        to specify the error message if the wrapped function fails.
    """
    def decorator(f, *,
                  err_msg='Critical failure computing report'):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                out = f(*args, **kwargs)
            except Exception:
                msg = datamodel.ReportMessage(
                    message=err_msg,
                    step='solarforecastarbiter.reports.main',
                    level='CRITICAL',
                    function=str(f)
                )
                raw = datamodel.RawReport(
                    pd.Timestamp.now(tz='UTC'), 'UTC', (), None,
                    (), (), (msg,))
                session.post_raw_report(report_id, raw, 'failed')
                raise
            else:
                return out
        return wrapper
    return decorator


def compute_report(access_token, report_id, base_url=None):
    """
    Create a raw report using data from API. Typically called as a task.
    Failures will attempt to post a message for the failure in an
    empty RawReport to the API.

    Parameters
    ----------
    session : :py:class:`solarforecastarbiter.api.APISession`
        API session for getting and posting data
    report_id : str
        ID of the report to fetch from the API and generate the raw
        report for

    Returns
    -------
    raw_report : :py:class:`solarforecastarbiter.datamodel.RawReport`
    """
    session = APISession(access_token, base_url=base_url)
    fail_wrapper = capture_report_failure(report_id, session)
    report = fail_wrapper(session.get_report, err_msg=(
        'Failed to retrieve report. Perhaps the report does not exist, '
        'the user does not have permission, or the connection failed.')
    )(report_id)
    data = fail_wrapper(get_data_for_report, err_msg=(
        'Failed to retrieve data for report which may indicate a lack '
        'of permissions or that an object does not exist.')
    )(session, report)
    raw_report = fail_wrapper(create_raw_report_from_data, err_msg=(
        'Unhandled exception when computing report.')
    )(report, data)
    fail_wrapper(session.post_raw_report, err_msg=(
        'Computation of report completed, but failed to upload result to '
        'the API.')
    )(report.report_id, raw_report)
    return raw_report
